# -*- encoding: utf-8 -*-
"""
hio.base Package
"""

from .tyming import Tymist, Tymee, Tymer
from .doing import Doist, doize, doify, Doer, DoDoer
from .filing import openFiler, Filer, FilerDoer
from .multidoing import Bosser, Crewer, TagDex
from .during import (Duror, openDuror, SuberBase, Suber, IoSuber, IoSetSuber,
                     DomSuberBase, DomSuber, DomIoSuber, DomIoSetSuber, Subery)



# -*- encoding: utf-8 -*-
"""hio.help.doming module

Support for Dataclass Object Mapper Dom classes
"""
from __future__ import annotations  # so type hints of classes get resolved later

import functools
from hashlib import blake2b
import json
import msgpack
import cbor2 as cbor

from typing import Any, Type, ClassVar
from collections.abc import Callable
from dataclasses import dataclass, astuple, asdict, fields, field, InitVar

from ..hioing import HierError
from .helping import NonStringIterable

# DOM Utilities dataclass utility classes


def dictify(val):
    """
    Returns a serializable dict represention of a dataclass.  If the dataclass
    contains a `_dictify method, use it instead of `asdict`

    Parameters:
         val the dataclass instance to turn into a dict.
    """
    dit = getattr(val, "_dictify", None)
    if callable(dit):
        return dit()

    return asdict(val)


def datify(cls, d):
    """Returns instance of dataclass cls converted from dict d. If the dataclass
    cls or any nested dataclasses contains a `_datify` method, the use it instead
    of default fieldtypes conversion.

    Parameters:
    cls is dataclass class
    d is dict
    """
    try:
        dat = getattr(cls, "_datify", None)
        if callable(dat):
            return dat(d)

        fieldtypes = {f.name: f.type for f in fields(cls)}
        return cls(**{f: datify(fieldtypes[f], d[f]) for f in d})  # recursive
    except Exception:  # Fields in dict d don't match dataclass or something else
        return d  # not a dataclass so end recursion and next level up will process


def registerify(cls):
    """Class Decorator to add cls as cls._registry entry for itself keyed by its
    own .__name__. Need class decorator so that class object is already created
    by registration time when decorator is applied
    """
    name = cls.__name__
    if name in cls._registry:
        raise HierError(f"RegDom subclass {name=} already registered.")
    cls._registry[name] = cls
    return cls


def namify(cls):
    """Class decorator for dataclass to populate ClassVar ._names with names
    of all fields in dataclass
    """
    cls._names = tuple(field.name for field in fields(cls))
    return cls



@dataclass(frozen=True)
class IceMapDom():
    """Base class for dataclasses (codexes) that support map syntax
    Adds support for dunder methods for map syntax dc[name].
    Converts exceptions from attribute syntax to raise map syntax when using
    map syntax.

    Enables dataclass instances to use Mapping item syntax except for __delattr__
    because dataclasses to not allow __delattr__ to be overridden

    A non-frozen dataclass may not be a subclass of a frozen dataclass neither
    may a frozen dataclass be a subclass of a frozen dataclass.

    Frozen dataclass do not allow change or upate of either field or non-field
    attributes after __init__ this includes in the  __post_init__ method.

    Class Methods:
        _fromdict(cls, d: dict): return dataclass converted from dict d

    Methods:
        __iter__():  asdict(self)
        __getitem__(self, name)  map like interface
        _asdict(): return self converted to dict
        _astuple(): return self converted to tuple

    """

    @classmethod
    def _fromdict(cls, d: dict):
        """returns instance of cls initialized from dict d """
        dom = datify(cls, d)
        if not isinstance(dom, cls):
            raise ValueError("Invalid dict={d} to datify as dataclass={cls}.")
        return dom


    def __iter__(self):
        return iter(self._asdict())


    def __getitem__(self, name):
        try:
            return getattr(self, name)
        except AttributeError as ex:
            raise IndexError(ex.args) from ex


    def _asdict(self):
        """Returns dict version of record"""
        return dictify(self)


    def _astuple(self):
        """Returns dict version of record"""
        return tuple(self._asdict().values())


@dataclass(frozen=True)
class IceRawDom(IceMapDom):
    """IceRawDom is frozen version of RawDom which is a subclass of IceMapDom
    that adds methods for converting to/from raw format serializations in JSON,
    CBOR, and MGPK. IceRawDom is a base class for such frozen dataclasses.
    The serialization process is to transform dataclass into
    dict and then serialize into raw. The deserialization process is to
    deserialize into dict and then convert to dataclass. This allows dataclass
    to stored in databases or to be sent over the wire in messages in raw format.

    Enables dataclass instances to use Mapping item syntax except for __delattr__
    because dataclasses to not allow __delattr__ to be overridden

    A non-frozen dataclass may not be a subclass of a frozen dataclass neither
    may a frozen dataclass be a subclass of a frozen dataclass.

    Frozen dataclass do not allow change or upate of either field or non-field
    attributes after __init__ this includes in the  __post_init__ method.


    Inherited Class Methods:
        _fromdict(cls, d: dict): return dataclass converted from dict d

    Class Methods:
        _fromjson(cls, s: str|bytes): return dataclass converted from json s
        _fromcbor(cls, s: bytes): return dataclass converted from cbor s
        _frommgpk(cls, s: bytes): return dataclass converted from mgpk s

    Inherited Methods:
        __iter__(self): asdict(self)
        __getitem__(self, name)  map like interface
        _asdict(self): return self converted to dict
        _astuple(): return self converted to tuple

    Methods:
        _asjson(self): return bytes self converted to json
        _ascbor(self): return bytes self converted to cbor
        _asmgpk(self): return bytes self converted to mgpk
    """

    @classmethod
    def _fromjson(cls, s: str | bytes):
        """returns instance of clas initialized from json str or bytes s """
        if hasattr(s, "decode"):  # bytes
            s = s.decode() # convert to str
        d = json.loads(s)  # convert to dict
        dom = datify(cls, d)
        if not isinstance(dom, cls):
            raise ValueError("Invalid dict={d} to datify as dataclass={cls}.")
        return dom


    @classmethod
    def _fromcbor(cls, s: bytes):
        """returns instance of clas initialized from cbor bytes or str s """
        d = cbor.loads(s)  # convert to dict
        dom = datify(cls, d)
        if not isinstance(dom, cls):
            raise ValueError("Invalid dict={d} to datify as dataclass={cls}.")
        return dom


    @classmethod
    def _frommgpk(cls, s: bytes):
        """returns instance of clas initialized from mgpk bytes or str s """
        d = msgpack.loads(s)  # convert to dict
        dom = datify(cls, d)
        if not isinstance(dom, cls):
            raise ValueError("Invalid dict={d} to datify as dataclass={cls}.")
        return dom


    def _asjson(self):
        """Returns json bytes version of record"""
        return json.dumps(self._asdict(),
                          separators=(",", ":"),
                          ensure_ascii=False).encode()  # ensure bytes


    def _ascbor(self):
        """Returns cbor bytes version of record"""
        return cbor.dumps(self._asdict())


    def _asmgpk(self):
        """Returns mgpk bytes version of record"""
        return msgpack.dumps(self._asdict())


@registerify
@dataclass(frozen=True)
class IceRegDom(IceRawDom):
    """IceRegDom is frozen version of RegDom which is a  base class for frozen Dom
    subclasses that have class registry as as non field ClassVar that holds
    mapping between class name and class object.
    Use the registerify decorator to populate the registry with a given dataclass.

    This allows registry based creation of instances from the class name.

    Enables dataclass instances to use Mapping item syntax except for __delattr__
    because dataclasses to not allow __delattr__ to be overridden

    A non-frozen dataclass may not be a subclass of a frozen dataclass neither
    may a frozen dataclass be a subclass of a frozen dataclass.

    Frozen dataclass do not allow change or upate of either field or non-field
    attributes after __init__ this includes in the  __post_init__ method.

    Non-Field Class Attributes:
        _registry (ClassVar[dict]): dict of subclasses keyed by class.__name__
            Assigned by @registerify decorator

    """
    _registry:  ClassVar[dict] = {}  # subclass registry


@namify
@registerify
@dataclass(frozen=True)
class IceTymeDom(IceRegDom):
    """IceTymeDom is frozen version of TymeDom which is a base class for frozen
    Dom dataclasses that are registered and support the cycle tyme stamping of
    named fields using non-field attributes. Note the only tyme stamp that can
    happen is at init since the dataclass is frozen.

    Use decorator registerify to populate ._registry
    Use deocrator namify to populate ._names which holds names of field attributes

    Enables dataclass instances to use Mapping item syntax except for __delattr__
    because dataclasses to not allow __delattr__ to be overridden

    A non-frozen dataclass may not be a subclass of a frozen dataclass neither
    may a frozen dataclass be a subclass of a frozen dataclass.

    Frozen dataclass do not allow change or upate of either field or non-field
    attributes after __init__ this includes in the  __post_init__ method.


    Inherited Non-Field Class Attributes:
        _registry (ClassVar[dict]): dict of subclasses keyed by class.__name__
            Assigned by @registerify decorator

    Non-Field Class Attributes:
        _names (ClassVar[tuple[str]|None]): tuple of field names for class
            Assigned by @namify decorator

    Properties:
        _tymth (None|Callable):  Emulates interface for non-frozen TymeDom
                                 Returns None
        _tyme (None|Float): Emulates interface for non-frozen TymeDom
                            Returns None
        _now (None|float): Emulates interface for non-frozen TymeDom
                           Returns None

    """
    _names: ClassVar[tuple[str]|None] = None  # Assigned in  __post_init__
    _tymth: InitVar[None|Callable] = None  # tymth closure not a field
    _tyme: InitVar[None|float] = None  # tyme of last update

    def __post_init__(self, _tymth, _tyme):
        """Emulates non-frozen TymeDom"""
        #self._tymth = _tymth
        #self._tyme = _tyme if _tyme is not None else self._now
        pass

    # cannot overwrite this method in frozen dataclass. so we use poperties
    # to emulate interface of TymeDom but always None
    #def __setattr__(self, name, value):
        #super().__setattr__(name, value)
        #if name in self._names:
            #super().__setattr__("_tyme", self._now)


    @property
    def _tymth(self):
        """Emulates interface for non-frozen TymeDom
        Returns:
            None
        """
        return None


    @property
    def _tyme(self):
        """Emulates interface for non-frozen TymeDom
        Returns:
            None
        """
        return None


    @property
    def _now(self):
        """Emulates interface for non-frozen TymeDom
        Returns:
            None
        """
        return None


    def _wind(self, tymth):
        """Emulates interface for non-frozen TymeDom
        """
        pass


@dataclass()
class MapDom():
    """MapDom is a subclass of IceMapDom and is a base class for dataclasses
    that support map syntax
    Adds support for dunder methods for map syntax dc[name].
    Converts exceptions from attribute syntax to raise map syntax when using
    map syntax.

    Enables dataclass instances to use Mapping item syntax except for __delattr__
    because dataclasses to not allow __delattr__ to be overridden

    A non-frozen dataclass may not be a subclass of a frozen dataclass neither
    may a frozen dataclass be a subclass of a frozen dataclass.

    Class Methods:
        _fromdict(cls, d: dict): return dataclass converted from dict d

    Methods:
        __iter__():  asdict(self)
        __getitem__(self, name)  map like interface
        __setitem__(self, name, value)  map like interface
        _update(*pa,**kwa): update attributes using dict like update syntax
        _asdict(): return self converted to dict
        _astuple(): return self converted to tuple
    """

    @classmethod
    def _fromdict(cls, d: dict):
        """returns instance of cls initialized from dict d """
        dom = datify(cls, d)
        if not isinstance(dom, cls):
            raise ValueError("Invalid dict={d} to datify as dataclass={cls}.")
        return dom


    def __iter__(self):
        return iter(self._asdict())


    def __getitem__(self, name):
        try:
            return getattr(self, name)
        except AttributeError as ex:
            raise IndexError(ex.args) from ex


    def __setitem__(self, name, value):
        try:
            return setattr(self, name, value)
        except AttributeError as ex:
            raise IndexError(ex.args) from ex



    def _update(self, *pa, **kwa):
        """Use item update syntax
        """
        if len(pa) > 1:
            raise TypeError(f"Expected 1 positional argument got {len(pa)}.")

        if pa:
            di = pa[0]
            if isinstance(di, Mapping):
                for k, v in di.items():
                    self[k] = v
            elif isinstance(di, NonStringIterable):
                for k, v in di:
                    self[k] = v
            else:
                raise TypeError(f"Expected Mapping or NonStringIterable got "
                                f"{type(di)}.")

        for k, v in kwa.items():
            self[k] = v


    def _asdict(self):
        """Returns dict version of record"""
        return dictify(self)


    def _astuple(self):
        """Returns tuple version of record"""
        return tuple(self._asdict().values())


def modify(mods: MapDom|None=None, klas: Type[MapDom]=MapDom)->Callable[..., Any]:
    """Wrapper with argument(s) that injects mods dataclass instance that must
    be a subclass of MapDom. When mods is provided it is injected as the keyword
    arg 'mods' into the wrapped function.
    If mods is None then an instance of klas is created and injected instead.
    This creates a lexical closure of this injection. The value of klas must
    also be a a MapDom subclass.

    Parameters:
        mods (WorkDom|None): default None. Instance of dataclass to be injected
        klas (Type[WorkDom]): defualt WorkDom. Class of dataclass to be injected
                             as default when mods not provided.

    If wrapped function call itself includes mods as an arg whose value is
    not None then does not inject. This allows override of a single call.
    Subsequent calls will resume using the lexical closure or the wrapped
    injected mods whichever was provided.

    Assumes wrapped function defines mods argument as a keyword only parameter
    using '*' such as:
       def f(name='box, over=None, *, mods=None):

    To use inline:
        mods = WorkDom(box=None, over=None, bepre='box', beidx=0)
        g = modify(mods)(f)

    Later calling g as:
        g(name="mid", over="top")
    Actually has mods inject as if called as:
        g(name="mid", over="top", mods=mods)

    Also can be used on a method not just a function.
       def m(self, name='box, over=None, *, mods=None):

    To use inline:
        mods = dict(box=None, over=None, bepre='box', beidx=0)
        m = modify(mods)(self.m)

    Later calling m as:
        m(name="mid", over="top")
    Actually has mods injectd as if called as:
        m(name="mid", over="top", mods=mods)  where self is also injected into method

    Since mods is a mutable collection i.e. dataclass, not an immutable object
    using @decorator syntax on could be problematic as the injected mods would
    be a lexical closure defined in the defining scope not the calling scope.
    Depends on what the use case it for it.

    Example:
        mods = WorkDom(box=None, over=None, bepre='box', beidx=0)
        @modify(mods=mods)
        def f(name="box), over=None, *, mods=None)

    Later calling f as:
        f(name="mid", over="top")
    Actually has mods injected as if called as:
        f(name="mid", over="top", mods=mods)

    But the mods in this case is from the defining scope,not the calling scope.

    Likewise passing in mods=None would result in a lexical closure of mods
    with default values initially that would be shared everywhere f() is called.

    When f() is called with an explicit mods such as f(mods=mods) then that
    call will use the passed in mods not the injected mods. This allows a per
    call override of the injected mods.

    """
    mods = mods if mods is not None else klas()  # lexical closure so not None
    def decorator(f):
        @functools.wraps(f)
        def inner(*pa, **kwa):
            if 'mods' not in kwa or kwa['mods'] is None:  # missing or None
                kwa.update(mods=mods)  # replace or add mods to kwa
            return f(*pa, **kwa)
        return inner
    return decorator


def modize(mods: dict|None=None) -> Callable[..., Any]:
    """Wrapper with argument(s) that injects dict or mods dataclass instance
    that must be a subclass of MapDom as lexical closure into the wrapped function
    as keyword parameter 'mods' .

    If wrapped function call itself includes mods as an arg whose value is
    not None then does not inject. This allows override of a single call.
    Subsequent calls will resume using the lexical closure or the wrapped
    injected mods whichever was provided.

    Assumes wrapped function defines mods argument as a keyword only parameter
    using '*' such as:
       def f(name='box, over=None, *, mods=None):

    To use inline:
        mods = dict(box=None, over=None, bepre='box', beidx=0)
        g = modize(mods)(f)

    Later calling g as:
        g(name="mid", over="top")
    Actually has mods injected as if called as:
        g(name="mid", over="top", mods=mods)

    If method then works as well.
       def f(self, name='box, over=None, *, mods=None):

    To use inline:
        mods = dict(box=None, over=None, bepre='box', beidx=0)
        f = modize(mods)(self.f)

    Later calling f as:
        f(name="mid", over="top")
    Actually has mods injected as if called as:
        f(name="mid", over="top", mods=mods)  the self is automaticall supplied

    Since mods is a mutable collection i.e. dict, not an immutable string then
    using it as decorator could be problematic as the mods would have lexical
    defining scope not calling scope.
    Which is ok if mods has lexical module scope intentionally.

    Example:
        mods = dict(box=None, over=None, bepre='box', beidx=0)

        @modize(mods=mods)
        def f(name="box), over=None, *, mods=None)

    Later calling as:
        f(name="mid", over="top")
    Actually has mods injected as if called as:
        f(name="mid", over="top", mods=mods)

    But the mods in this case is from the defining scope not the calling scope.

    Likewise passing in mods=None would result in a lexical closure of mods
    with default values initially that would be shared everywhere f() is called
    with whatever values each call changes in the lexical closure of mods.
    """
    mods = mods if mods is not None else {}  # {} in lexical closure when None
    def decorator(f):
        @functools.wraps(f)
        def inner(*pa, **kwa):
            if 'mods' not in kwa or kwa['mods'] is None:  # missing or None
                kwa.update(mods=mods)  # replace or add works to kwa
            return f(*pa, **kwa)
        return inner
    return decorator


@dataclass
class RawDom(MapDom):
    """RawDom is subclass of MapDom that adds methods for converting to/from
    raw format serializations in JSON, CBOR, and MGPK. RawDom is a base class
    for such dataclasses. The serialization process is to transform dataclass into
    dict and then serialize into raw. The deserialization process is to
    deserialize into dict and then convert to dataclass. This allows dataclass
    to stored in databases or to be sent over the wire in messages in raw format.

    Enables dataclass instances to use Mapping item syntax except for __delattr__
    because dataclasses to not allow __delattr__ to be overridden

    A non-frozen dataclass may not be a subclass of a frozen dataclass neither
    may a frozen dataclass be a subclass of a frozen dataclass.


    Inherited Class Methods:
        _fromdict(cls, d: dict): return dataclass converted from dict d

    Class Methods:
        _fromjson(cls, s: str|bytes): return dataclass converted from json s
        _fromcbor(cls, s: bytes): return dataclass converted from cbor s
        _frommgpk(cls, s: bytes): return dataclass converted from mgpk s

    Inherited Methods:
        __getitem__(self, name)  map like interface
        __setitem__(self, name, value)  map like interface
        __iter__(self): asdict(self)
        _asdict(self): return self converted to dict
        _astuple(): return self converted to tuple
        _update(*pa,**kwa): update attributes using dict like update syntax

    Methods:
        _asjson(self): return bytes self converted to json
        _ascbor(self): return bytes self converted to cbor
        _asmgpk(self): return bytes self converted to mgpk
    """

    @classmethod
    def _fromjson(cls, s: str | bytes):
        """returns instance of clas initialized from json str or bytes s """
        if hasattr(s, "decode"):  # bytes
            s = s.decode() # convert to str
        d = json.loads(s)  # convert to dict
        dom = datify(cls, d)
        if not isinstance(dom, cls):
            raise ValueError("Invalid dict={d} to datify as dataclass={cls}.")
        return dom


    @classmethod
    def _fromcbor(cls, s: bytes):
        """returns instance of clas initialized from cbor bytes or str s """
        d = cbor.loads(s)  # convert to dict
        dom = datify(cls, d)
        if not isinstance(dom, cls):
            raise ValueError("Invalid dict={d} to datify as dataclass={cls}.")
        return dom


    @classmethod
    def _frommgpk(cls, s: bytes):
        """returns instance of clas initialized from mgpk bytes or str s """
        d = msgpack.loads(s)  # convert to dict
        dom = datify(cls, d)
        if not isinstance(dom, cls):
            raise ValueError("Invalid dict={d} to datify as dataclass={cls}.")
        return dom


    def _asjson(self):
        """Returns json bytes version of record"""
        return json.dumps(self._asdict(),
                          separators=(",", ":"),
                          ensure_ascii=False).encode()  # ensure bytes


    def _ascbor(self):
        """Returns cbor bytes version of record"""
        return cbor.dumps(self._asdict())


    def _asmgpk(self):
        """Returns mgpk bytes version of record"""
        return msgpack.dumps(self._asdict())



@registerify
@dataclass
class RegDom(RawDom):
    """RegDom is base class for Dom subclasses that have class registry as
    as non field ClassVar that holds mapping between class name and class object.
    Use the registerify decorator to populate the registry with a given dataclass.

    This allows registry based creation of instances from the class name.

    Enables dataclass instances to use Mapping item syntax except for __delattr__
    because dataclasses to not allow __delattr__ to be overridden

    A non-frozen dataclass may not be a subclass of a frozen dataclass neither
    may a frozen dataclass be a subclass of a frozen dataclass.

    Non-Field Class Attributes:
        _registry (ClassVar[dict]): dict of subclasses keyed by class.__name__
            Assigned by @registerify decorator

    """
    _registry:  ClassVar[dict] = {}  # subclass registry


    def __hash__(self):
        """Define hash so can work with ordered_set
        __hash__ is not inheritable in dataclasses so must be explicitly defined
        in every subclass
        """
        return hash((self.__class__.__name__,) + self._astuple())  # almost same as __eq__


@namify
@registerify
@dataclass
class TymeDom(RegDom):
    """TymeDom is base class for Dom dataclasses that are registered and support
    the cycle tyme stamping of updates to named fields using non-field attributes.
    Use decorator registerify to populate ._registry
    Use deocrator namify to populate ._names which holds names of field attributes

    Supports the cycle tyme stamping of updates to named fields using non-field attributes.

    This allows registry based creation of instances from the class name.

    Enables dataclass instances to use Mapping item syntax except for __delattr__
    because dataclasses to not allow __delattr__ to be overridden

    A non-frozen dataclass may not be a subclass of a frozen dataclass neither
    may a frozen dataclass be a subclass of a frozen dataclass.

    Inherited Non-Field Class Attributes:
        _registry (ClassVar[dict]): dict of subclasses keyed by class.__name__
            Assigned by @registerify decorator

    Non-Field Class Attributes:
        _names (ClassVar[tuple[str]|None]): tuple of field names for class
            Assigned by @namify decorator

    Non-Field Attributes:
        _tymth (None|Callable): function wrapper closure returned by
            Tymist.tymen() method. When .tymth is called it returns associated
            Tymist.tyme. Provides injected dependency on Tymist cycle tyme base.
            None means not assigned yet.
            Use ._wind method to assign ._tymth after init of bag.
        _tyme (None|Float): cycle tyme of last update of a bag field.
            None means either ._tymth as not yet been assigned or this bag's
            fields have not yet been updated.

    Properties:
        _now (None|float): current tyme given by ._tymth if not None.

    """
    _names: ClassVar[tuple[str]|None] = None  # Assigned in  __post_init__
    _tymth: InitVar[None|Callable] = None  # tymth closure not a field
    _tyme: InitVar[None|float] = None  # tyme of last update

    def __post_init__(self, _tymth, _tyme):  # after init so fields already setup
        self._tymth = _tymth
        self._tyme = _tyme


    def __hash__(self):
        """Define hash so can work with ordered_set
        __hash__ is not inheritable in dataclasses so must be explicitly defined
        in every subclass
        """
        return hash(self._astuple())  # same as default dataclass __eq__()


    def __setattr__(self, name, value):
        super().__setattr__(name, value)
        if name in self._names:
            super().__setattr__("_tyme", self._now)


    @property
    def _now(self):
        """Gets current tyme from injected ._tymth closure from Tymist.
        tyme is float cycle time in seconds
        Returns:
            _now (float | None): tyme from self.tymth() when wound else None
        """
        return self._tymth() if self._tymth else None


    def _wind(self, tymth):
        """
        Inject new tymist.tymth as new ._tymth. Changes tymist.tyme base.
        Override in subclasses to update any dependencies on a change in
        tymist.tymth base
        """
        self._tymth = tymth




# -*- encoding: utf-8 -*-
"""hio.base.hier.durqing Module

Provides durable  queue for hold hierarchical actions

"""
from __future__ import annotations  # so type hints of classes get resolved later

from collections import deque
from typing import Any

from hio import HierError
from ...help import RegDom, IceRegDom, NonStringIterable

class Durq():
    """Durq (durable queue) class when injected with
    .sdb and .key will store its ordered list durably and allow access as a FIFO
    queue

    Properties
        stale (bool): True means in-memory and durable on disk not synced
                     False means in-memory and durable on disk synced
        durable (bool): True means ._sdb and ._key and ._sdb.db and
                                .sdb.db.opened are not None
                        False otherwise


    Hidden:
       _deq (deque):  in-memory cache as deque
       _sdb (DomIoSuber): instance of durable store
       _key (str): into .sdb
       _stale (bool): for .stale property:
                       True means in-memory and durable on disk not synced
                       False means in-memory and durable on disk synced

    """
    def __init__(self, *pa):
        """Initialize instance

        Parameters:
           pa[0] (NonStringeIteralble[RegDom]): instances to preload self._deq

        """
        self._deq = deque()
        self._stale = True

        self._sdb = None
        self._key = None

        if len(pa) > 1:
            raise TypeError(f"Expected 1 positional argument got {len(pa)}.")

        if pa:
            di = pa[0]
            if isinstance(di, NonStringIterable):
                self.extend(di)

            else:
                raise TypeError(f"Expected NonStringIterable got "
                                f"{type(di)}.")


    def __repr__(self):
        """Custom repr for Durq"""
        return (f"Durq({repr(list(self._deq))})")


    def __iter__(self):
        """Makes iterator out of self by returning iterable ._deq"""
        return iter(self._deq)

    def __len__(self):
        """Supports len()"""
        return len(self._deq)


    @property
    def stale(self):
        """Getter for ._stale

        Returns:
            stale (bool): True means in-memory and durable on disk not synced
                         False means in-memory and durable on disk synced
        """
        return self._stale

    @property
    def durable(self):
        """Property durable True when durable subdb injected and opened.

        Returns:
            durable (bool): True means ._sdb and ._key and ._sdb.db and
                                .sdb.db.opened are not None
                            False otherwise
        """
        return (self._sdb is not None and self._key is not None and self._sdb.db
                and self._sdb.db.opened)


    def extend(self, vals: NonStringIterable[RegDom|IceRegDom]):
        """Extend ._deq with vals
        Peforms equivalent operation on durable .sdb at .key if any

        """
        if not vals:
            return False

        for val in vals:
            if not isinstance(val, (RegDom, IceRegDom)):
                raise HierError(f"Expected RegDom instance got {val}")

        self._deq.extend(vals)
        if self.put(vals) is False:  # durable but put failed
            raise HierError(f"Mismatch between cache and durable at "
                            f"key={self._key}")
        return True


    def push(self, val: RegDom|IceRegDom):
        """If not None, add val to last in. Otherwise ignore
        Peforms equivalent operation on durable .sdb at .key if any

        Parameters:
            val (RegDom): element to be appended to deck (deque)
        """
        if val is not None:
            if not isinstance(val, (RegDom, IceRegDom)):
                raise HierError(f"Expected RegDom instance got {val}")
            self._deq.append(val)
            result = self.add(val)
            if result == False:  # durable but not added
                raise HierError(f"Mismatch between cache and durable at "
                                f"key={self._key}")
            return True
        return False


    def pull(self, emptive=True):
        """Remove and return first in value
        If empty and emptive return None else raise IndexError

        Peforms equivalent operation on durable .sdb at .key if any

        Parameters:
            emptive (bool): True means return None instead of raise IndexError
               when attempt to pull
               False means normal behavior of deque
        """
        try:
            val = self._deq.popleft()
        except IndexError:  # empty
            if self.durable and self.pop() is not None:
                raise HierError(f"Mismatch between cache and durable at "
                                f"key={self._key}")
            if not emptive:
                raise
            return None
        else:  # successfully popped so pop from ._sdb
            if self.durable and self.pop() is None:  # not popped from ._sdb
                raise HierError(f"Mismatch between cache and durable at "
                                f"key={self._key}")

        return val  # value to return


    def clear(self):
        """Clear all values from deque
        Peforms equivalent operation on durable .sdb at .key if any

        """
        prior = len(self._deq)
        self._deq.clear()
        unique = prior > 0
        if not unique:
            return False

        if self.rem() == False:
            raise HierError(f"Mismatch between cache and durable at "
                            f"key={self._key}")
        return True


    def count(self, value):
        """Returns count of entries in ._deq equal to value
        """
        return(self._deq.count(value))  # counts matching values



    def put(self, vals: NonStringIterable[RegDom|IceRegDom]):
        """Put (append) vals to .sdb at .key if any"""
        if self.durable:
            self._stale = False
            return self._sdb.put(keys=self._key, vals=vals)
        return None


    def add(self, val):
        """Add value to .sdb at .key if any"""
        if self.durable:
            self._stale = False
            return self._sdb.add(keys=self._key, val=val)
        return None


    def pop(self):
        """Pop value from .sdb at .key if any"""
        if self.durable:
            return self._sdb.pop(keys=self._key)
        return None


    def rem(self):
        """Remove all values from .sdb at .key if any"""
        if self.durable:
            return self._sdb.rem(keys=self._key)
        return None


    def cnt(self):
        """Count all values in .sdb at .key if any"""
        if self.durable:
            return self._sdb.cnt(keys=self._key)
        return None


    def pin(self):
        """Pins all of ._deq to ._sdb at ._key if any.
        Sets ._stale to False on success
        """
        if self.durable:
            vals = [val for val in self._deq]
            self._sdb.pin(self._key, vals)  # pin sdb._ser to ._deq vals
            self._stale = False
            return True
        return None


    def sync(self, force=False):
        """Syncs by checking exists durable copy in ._sdb at ._key. If so
        and ._stale or force then reads own data from ._sdb at ._key if any.
        If not attempts to write to ._sdb at ._key

        Parameters:
            force (bool): True means force read even if not ._stale
                          Flase means do not force read
        """
        if self.durable and (self.stale or force):
            if self._sdb.cnt(self._key):  # not empty
                self._deq.clear()
                self._deq.extend(self._sdb.getIter(self._key))
                self._stale = False
                return True

            else: # empty
                return self.pin()

        return None




# -*- encoding: utf-8 -*-
"""hio.help.during module

Support for Durable storage using LMDB
"""

from __future__ import annotations  # so type hints of classes get resolved later

import os
import platform
import stat
import tempfile
from contextlib import contextmanager

import lmdb
from ordered_set import OrderedSet as oset

import hio

from hio import HierError
from .doing import Doer
from .filing import Filer
from ..help import isNonStringIterable, RegDom, IceRegDom



class Duror(Filer):
    """Manages an an LMDB database directory and environment.

    Inherited Class Attributes:  (see Filer)
        HeadDirPath (str): default abs dir path head such as "/usr/local/var"
        TailDirPath (str): default rel dir path tail when using head
        CleanTailDirPath (str): default rel dir path tail when creating clean
        AltHeadDirPath (str): default alt dir path head such as  "~"
                              as fallback when desired head not permitted.
        AltTailDirPath (str): default alt rel dir path tail as fallback
                              when using alt head.
        AltCleanTailDirPath (str): default alt rel path tail when creating clean
        TempHeadDir (str): default temp abs dir path head such as "/tmp"
        TempPrefix (str): default rel dir path prefix when using temp head
        TempSuffix (str): default rel dir path suffix when using temp head and tail
        Perm (int): explicit default octal perms such as 0o1700
        Mode (str): open mode such as "r+"
        Fext (str): default file extension such as "text" for "fname.text"


    Inherited Attributes:  (see Filer)
        name (str): unique path component at end of path for directory or file
        base (str): another unique path component inserted before name
        temp (bool): True means use TempHeadDir in /tmp directory
        headDirPath (str): head directory path
        path (str | None):  full directory or file path once created else None
        perm (int):  octal OS permissions for path directory and/or file
        filed (bool): True means .path ends in file.
                      False means .path ends in directory
        extensioned (bool): When not filed:
                                True means ensure .path ends with fext
                                False means do not ensure .path ends with fext
        mode (str): file open mode if filed
        fext (str): file extension if filed
        file (File | None): File instance when filed and created.
        opened (bool): True means directory created and if filed then file
                is opened. False otherwise


    Attributes:
        env (lmdb.env): LMDB main (super) database environment
        readonly (bool): True means open LMDB env as readonly

    Properties:
        version

    File/Directory Creation Mode Notes:
        .Perm provides default restricted access permissions to directory and/or files
        stat.S_ISVTX | stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR
        0o1700==960

        stat.S_ISVTX  is Sticky bit. When this bit is set on a directory it means
            that a file in that directory can be renamed or deleted only by the
            owner of the file, by the owner of the directory, or by a privileged process.
            When this bit is set on a file it means nothing
        stat.S_IRUSR Owner has read permission.
        stat.S_IWUSR Owner has write permission.
        stat.S_IXUSR Owner has execute permission.
    """
    HeadDirPath = os.path.join(os.path.sep, "usr", "local", "var")  # default in /usr/local/var
    TailDirPath = os.path.join("hio", "db")
    CleanTailDirPath = os.path.join("hio", "clean", "db")
    AltHeadDirPath = os.path.expanduser("~")  # put in ~ as fallback when desired not permitted
    AltTailDirPath = os.path.join(".hio", "db")
    AltCleanTailDirPath = os.path.join(".hio", "clean", "db")
    TempHeadDir = (os.path.join(os.path.sep, "tmp")
                   if platform.system() == "Darwin" else tempfile.gettempdir())
    TempPrefix = "hio_lmdb_"
    TempSuffix = "_test"
    Perm = stat.S_ISVTX | stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR  # 0o1700==960
    MaxNamedDBs = 16
    MapSize = 104857600
    SuffixSize = 32  # does not include separator
    MaxSuffix = int("f"*(SuffixSize), 16)


    def __init__(self, readonly=False, **kwa):
        """Setup main database directory at .dirpath.
        Create main database environment at .env using .path.

        Inherited Parameters: (see Filer)
            name (str): Unique identifier of file. Unique directory path name
                differentiator directory/file
                When system employs more than one keri installation, name allows
                differentiating each instance by name
            base (str): optional directory path segment inserted before name
                that allows further differentation with a hierarchy. "" means
                optional.
            temp (bool): assign to .temp
                True then open in temporary directory, clear on close
                Otherwise then open persistent directory, do not clear on close
            headDirPath (str): optional head directory pathname for main database
                Default .HeadDirPath
            perm (int): optional numeric os dir permissions for database
                directory and database files. Default .DirMode
            reopen (bool): True (re)open with this init
                           False not (re)open with this init but later (default)
            clear (bool): True means remove directory upon close when reopening
                          False means do not remove directory upon close when reopening
            reuse (bool): True means reuse self.path if already exists
                          False means do not reuse but remake self.path
            clean (bool): True means path uses clean tail variant
                             False means path uses normal tail variant
            filed (bool): True means .path is file path not directory path
                          False means .path is directiory path not file path
            extensioned (bool): When not filed:
                                True means ensure .path ends with fext
                                False means do not ensure .path ends with fext
            mode (str): File open mode when filed
            fext (str): File extension when filed or extensioned

        Parameters:
            readonly (bool): True means open database in readonly mode
                                False means open database in read/write mode

        """

        self.env = None
        self._version = None
        self.readonly = True if readonly else False
        super(Duror, self).__init__(**kwa)


    def reopen(self, readonly=False, **kwa):
        """Open if closed or close and reopen if opened or create and open if not
        if not preexistent, directory path for lmdb at .path and then
        Open lmdb and assign to .env

        Inherited Parameters:  (see Filer)
            temp (bool): assign to .temp
                         True means open in temporary directory, clear on close
                         False means open persistent directory, do not clear on close
            headDirPath (str): optional head directory pathname for main database
                               Default .HeadDirpath
            perm (int): optional numeric os dir permissions for database
                         directory and database files. Default .Perm
            clear (bool): True means remove directory upon close
                             False means do not remove directory upon close
            reuse (bool): True means reuse self.path if already exists
                             False means do not reuse but remake self.path
            clean (bool): True means path uses clean tail variant
                             False means path uses normal tail variant
            mode (str): file open mode when .filed
            fext (str): File extension when .filed

        Parameters:
            readonly (bool): True means open database in readonly mode
                                False means open database in read/write mode
        """
        exists = self.exists(name=self.name, base=self.base)
        opened = super(Duror, self).reopen(**kwa)
        if readonly is not None:
            self.readonly = readonly

        # open lmdb major database instance
        # creates files data.mdb and lock.mdb in .dbDirPath
        self.env = lmdb.open(self.path,
                             max_dbs=self.MaxNamedDBs,
                             map_size=self.MapSize,
                             mode=self.perm,
                             readonly=self.readonly)

        self.opened = True if opened and self.env else False

        if self.opened and not self.readonly and (not exists or self.temp):
            self.version = hio.__version__

        return self.opened

    @property
    def version(self):
        """ Return the version of database stored in __version__ key.

        This value is read through cached in memory

        Returns:
            version (str): semver string
                the version of the database or None if not set in the database

        """
        if self._version is None:
            self._version = self.getVer()

        return self._version

    @version.setter
    def version(self, version):
        """Set the version of the database in memory and in the __version__ key

        Parameters:
            version (str): The new semver formatted version of the database

        """
        if hasattr(version, "decode"):
            version = version.decode("utf-8")  # convert bytes to str

        self._version = version
        self.setVer(self._version)


    def close(self, clear=False):
        """Close lmdb at .env
        if clear or .temp then remove lmdb directory at .path

        Parameters:
           clear (bool), True means clear lmdb directory after close
                         Otherwise do not clear after close
        """
        if self.env:
            try:
                self.env.close()
            except:
                pass

        self.env = None

        return super(Duror, self).close(clear=clear)


    def getVer(self):
        """Returns the value of the the semver formatted version in the
        __version__ key in this database

        Returns:
            str: semver formatted version of the database

        """
        with self.env.begin() as txn:
            cursor = txn.cursor()
            version = cursor.get(b'__version__')
            return version.decode("utf-8") if version is not None else None


    def setVer(self, version):
        """ Set the version of the database in the __version__ key

        Parameters:
            version (str): The new semver formatted version str of the database

        """
        if hasattr(version, "encode"):
            version = version.encode("utf-8")  # convert str to bytes

        with self.env.begin(write=True) as txn:
            cursor = txn.cursor()
            cursor.replace(b'__version__', version)


    # For subdbs with no duplicate values allowed at each key. (dupsort==False)
    def putVal(self, sdb, key, val):
        """Write serialized bytes val to location key in db (subdb)
        Does not overwrite.
        Returns:
            result (bool):  True if val successfully written
                            False if val at key already exitss

        Parameters:
            db (lmdb._Database): opened named subdb with dupsort=False
            key (bytes): within subdb's keyspace
            val (bytes):  to be written at key
        """
        with self.env.begin(db=sdb, write=True, buffers=True) as txn:
            try:
                return (txn.put(key, val, overwrite=False))
            except lmdb.BadValsizeError as ex:
                raise KeyError(f"Key: `{key}` is either empty, too big (for lmdb),"
                               " or wrong DUPFIXED size.") from ex


    def pinVal(self, sdb, key, val):
        """Write serialized bytes val to location key in db
        Overwrites existing val if any
        Returns:
            result (bool): True If val successfully written
                           False otherwise

        Parameters:
            db (lmdb._Database): opened named subdb with dupsort=False
            key (bytes): within subdb's keyspace
            val (bytes):  to be written at key
        """
        with self.env.begin(db=sdb, write=True, buffers=True) as txn:
            try:
                return (txn.put(key, val))
            except lmdb.BadValsizeError as ex:
                raise KeyError(f"Key: `{key}` is either empty, too big (for lmdb),"
                               " or wrong DUPFIXED size.") from ex


    def getVal(self, sdb, key):
        """Get val at key in db

        Returns:
            val (bytes):  None if no entry at key

        Parameters:
            db (lmdb._Database): opened named subdb with dupsort=False
            key (bytes): within subdb's keyspace

        """
        with self.env.begin(db=sdb, write=False, buffers=True) as txn:
            try:
                val = txn.get(key)
                return (bytes(val) if val is not None else None)
            except lmdb.BadValsizeError as ex:
                raise KeyError(f"Key: `{key}` is either empty, too big (for lmdb),"
                               " or wrong DUPFIXED size.") from ex


    def remVal(self, sdb, key):
        """Removes value at key in db.

        Returns:
            result (bool): True If key exists in database
                           False otherwise

        Parameters:
            db (lmdb._Database): opened named subdb with dupsort=False
            key (bytes): within subdb's keyspace
        """
        with self.env.begin(db=sdb, write=True, buffers=True) as txn:
            try:
                return (txn.delete(key))
            except lmdb.BadValsizeError as ex:
                raise KeyError(f"Key: `{key}` is either empty, too big (for lmdb),"
                               " or wrong DUPFIXED size.") from ex


    def cntVals(self, sdb):
        """Counts entries in subdb db

        Returns:
            count (int): number of entries in subdb db, or zero otherwise

        Parameters:
            db (lmdb._Database): opened named subdb with dupsort=False
        """
        with self.env.begin(db=sdb, write=False, buffers=True) as txn:
            cursor = txn.cursor()
            count = 0
            for _, _ in cursor:
                count += 1
            return count


    def getTopItemIter(self, sdb, top=b''):
        """Iterates over branch of subdb db rooted at top key

        Returns:
            items (Iterator): of (full key, val) tuples over a branch of the db
                given by top key where: full key is full database key for val
                not truncated top key

        Raises StopIteration Error when empty.

        Parameters:
            db (lmdb._Database): instance of named sub db with dupsort==False
            top (bytes): truncated top key, a key space prefix to get all the items
                        from multiple branches of the key space. If top key is
                        empty then gets all items in database.
                        In Python str.startswith('') always returns True so if branch
                        key is empty string it matches all keys in db with startswith.
        """
        with self.env.begin(db=sdb, write=False, buffers=True) as txn:
            cursor = txn.cursor()
            if cursor.set_range(top):  # move to val at key >= key if any
                for ckey, cval in cursor.iternext():  # get key, val at cursor
                    ckey = bytes(ckey)
                    if not ckey.startswith(top): #  prev entry if any last in branch
                        break  # done
                    yield (ckey, cval)  # another entry in branch startswith key
            return  # done raises StopIteration


    def remTopVals(self, sdb, top=b''):
        """Removes all values in branch of db given top key.

        Returns:
            result (bool): True if values were deleted at key.
                           False otherwise if no values at key

        Parameters:
            db (lmdb._Database): instance of named sub db with dupsort==False
            top (bytes): truncated top key, a key space prefix to get all the items
                        from multiple branches of the key space. If top key is
                        empty then deletes all items in database

        Raises StopIteration Error when empty.

        """
        # when deleting can't use cursor.iternext() because the cursor advances
        # twice (skips one) once for iternext and once for delete.
        with self.env.begin(db=sdb, write=True, buffers=True) as txn:
            result = False
            cursor = txn.cursor()
            if cursor.set_range(top):  # move to val at key >= key if any
                ckey, cval = cursor.item()
                while ckey:  # end of database key == b''
                    ckey = bytes(ckey)
                    if not ckey.startswith(top): #  prev entry if any last in branch
                        break  # done
                    result = cursor.delete() or result # delete moves cursor to next item
                    ckey, cval = cursor.item()  # cursor now at next item after deleted
            return result


    # insertion ordered key suffix and unsuffix
    @staticmethod
    def suffix(key: bytes|str|memoryview, ion: int, *, sep: bytes|str=b'.'):
        """
        Returns:
           iokey (bytes): actual DB key after concatenating suffix as hex version
           of insertion ordering ordinal int ion using separator sep.

        Parameters:
            key (bytes|str|memoryview): apparent effective database key (unsuffixed)
            ion (int)): insertion ordered ordinal numberfor set of vals
            sep (bytes): separator character(s) for concatenating suffix
        """
        if isinstance(key, memoryview):
            key = bytes(key)
        elif hasattr(key, "encode"):
            key = key.encode()  # encode str to bytes
        if hasattr(sep, "encode"):
            sep = sep.encode()
        ion =  b"%032x" % ion
        return sep.join((key, ion))

    @staticmethod
    def unsuffix(iokey: bytes|str|memoryview, *, sep: bytes|str=b'.'):
        """
        Returns:
           result (tuple): (bytes: key, int: ion) from splitting iokey at rightmost
            separator sep.
            Where key is bytes apparent effective DB key
            Where ion is int insertion orderered ordinal number converted from
                stripped off hex suffix

        Parameters:
            iokey (bytes|str|memoryview): actual database key suffixed
            sep (bytes): separator character(s) for splitting suffix
        """
        if isinstance(iokey, memoryview):
            iokey = bytes(iokey)
        elif hasattr(iokey, "encode"):
            iokey = iokey.encode()  # encode str to bytes
        if hasattr(sep, "encode"):
            sep = sep.encode()
        key, ion = iokey.rsplit(sep=sep, maxsplit=1)
        ion = int(ion, 16)
        return (key, ion)

    #Io and IoSet get methods work the same since IoSet operations do not dedup
    # on get, cnt, or del all only on add, put, set or del given val.
    # So common get, cnt, del all operations here and the add, put, set, del one
    # are provided below for IoSet after Io
    # Common methods to Io and IoSet
    def getIoValFirst(self, sdb, key, *, ion=0, sep=b'.'):
        """Gets first of the insertion ordered set of values at key,
        None if no entry.

        Returns:
            val (bytes|None): first val at same apparent effective key
                        whose iokey >= iokey from ion
                        Uses hidden ordinal key suffix for insertion ordering.
                        The suffix is appended and stripped transparently.
                        None if no entry at key

        Parameters:
            db (lmdb._Database): instance of named sub db with dupsort==False
            key (bytes): Apparent effective key
            ion (int): starting ordinal value, default 0

        """
        with self.env.begin(db=sdb, write=False, buffers=True) as txn:
            iokey = self.suffix(key, ion, sep=sep)  # start ion_th value for key zeroth default
            cursor = txn.cursor()
            if cursor.set_range(iokey):  # move to val at key >= iokey if any
                iokey, cval = cursor.item()  # get iokey, cval at cursor
                ckey, cion = self.unsuffix(iokey, sep=sep)
                if ckey == key:  # first entry for key >= iokey at ion
                    return bytes(cval)
            return None  # no item at key


    def getIoValLast(self, sdb, key, *, sep=b'.'):
        """Gets last value (last in) of insertion ordered set values at key

        Returns:
            val (bytes|None): last added empty at apparent effective
                    key if any, otherwise None if no entry

        Uses hidden ordinal key suffix for insertion ordering.
            The suffix is appended and stripped transparently.

        Parameters:
            db (lmdb._Database): instance of named sub db with dupsort==False
            key (bytes): Apparent effective key
        """
        val = None
        ion = None  # no last value
        iokey = self.suffix(key, ion=self.MaxSuffix, sep=sep)  # make iokey at max and walk back
        with self.env.begin(db=sdb, write=False, buffers=True) as txn:
            cursor = txn.cursor()  # create cursor to walk back
            if not cursor.set_range(iokey):  # max is past end of database
                # Three possibilities for max past end of database
                # 1. last entry in db is for same key
                # 2. last entry in db is for other key before key
                # 3. database is empty
                if cursor.last():  # not 3. empty db, so either 1. or 2.
                    ckey, cion = self.unsuffix(cursor.key(), sep=sep)
                    if ckey == key:  # 1. last is last entry for same key
                        ion = cion  # so set ion to cion
            else:  # max is not past end of database
                # Two possibilities for max not past end of databseso
                # 1. cursor at max entry at key
                # 2. other key after key with entry in database
                ckey, cion = self.unsuffix(cursor.key(), sep=sep)
                if ckey == key:  # 1. last entry for key is already at max
                    ion = cion
                else:  # 2. other key after key so backup one entry
                    # Two possibilities: 1. no prior entry 2. prior entry
                    if cursor.prev():  # prev entry, maybe same or earlier pre
                        # 2. prior entry with two possiblities:
                        # 1. same key
                        # 2. other key before key
                        ckey, cion = self.unsuffix(cursor.key(), sep=sep)
                        if ckey == key:  # prior (last) entry at key
                            ion = cion  # so set ion to the cion

            if ion is not None:
                iokey = self.suffix(key, ion=ion, sep=sep)
                val = bytes(cursor.get(iokey))

            return val


    def getIoVals(self, sdb, key, *, ion=0, sep=b'.'):
        """Gets list of all the insertion ordered set of values at key

        Returns:
            vals (list[bytes]): the insertion ordered set of values at same apparent
                           effective key.
                          Uses hidden ordinal key suffix for insertion ordering.
                          The suffix is appended and stripped transparently.

        Parameters:
            db (lmdb._Database): instance of named sub db with dupsort==False
            key (bytes): Apparent effective key
            ion (int): starting ordinal value, default 0

        """
        with self.env.begin(db=sdb, write=False, buffers=True) as txn:
            vals = []
            iokey = self.suffix(key, ion, sep=sep)  # start ion th value for key zeroth default
            cursor = txn.cursor()
            if cursor.set_range(iokey):  # move to val at key >= iokey if any
                for iokey, cval in cursor.iternext():  # get iokey, val at cursor
                    ckey, cion = self.unsuffix(iokey, sep=sep)
                    if ckey != key:  # prev entry if any was the last entry for key
                        break  # done
                    vals.append(bytes(cval))  # another entry at key
            return vals


    def getIoValsIter(self, sdb, key, *, ion=0, sep=b'.'):
        """Gets Iterator of all the insertion ordered set of values at key

        Returns:
            ioset (Iterator): iterator over insertion ordered set of values
                              at same apparent effective key.
                              Uses hidden ordinal key suffix for insertion ordering.
                              The suffix is appended and stripped transparently.

        Raises StopIteration Error when empty.

        Parameters:
            db (lmdb._Database): instance of named sub db with dupsort==False
            key (bytes): Apparent effective key
            ion (int): starting ordinal value, default 0
        """
        with self.env.begin(db=sdb, write=False, buffers=True) as txn:
            iokey = self.suffix(key, ion, sep=sep)  # start ion th value for key zeroth default
            cursor = txn.cursor()
            if cursor.set_range(iokey):  # move to val at key >= iokey if any
                for iokey, cval in cursor.iternext():  # get key, val at cursor
                    ckey, cion = self.unsuffix(iokey, sep=sep)
                    if ckey != key: #  prev entry if any was the last entry for key
                        break  # done
                    yield (cval)  # another entry at key
            return  # done raises StopIteration


    def popIoVal(self, sdb, key, *, ion=0, sep=b'.'):
        """Pops first of the insertion ordered set of values at key.
        None if no entry. Pop deletes the returned entry.

        Returns:
            val (bytes|None): first val at same apparent effective key
                        whose iokey >= iokey from ion
                        Uses hidden ordinal key suffix for insertion ordering.
                        The suffix is appended and stripped transparently.
                        None if no entry.

        Parameters:
            db (lmdb._Database): instance of named sub db with dupsort==False
            key (bytes): Apparent effective key
            ion (int): starting ordinal value, default 0

        """
        with self.env.begin(db=sdb, write=True, buffers=True) as txn:
            iokey = self.suffix(key, ion, sep=sep)  # start ion_th value for key zeroth default
            cursor = txn.cursor()
            if cursor.set_range(iokey):  # move to val at key >= iokey if any
                iokey, cval = cursor.item()  # get iokey, cval at cursor
                ckey, cion = self.unsuffix(iokey, sep=sep)
                if ckey == key:  # first entry for key >= iokey at ion
                    val = bytes(cval) # make copy so not deleted
                    cursor.delete()
                    return val
            return None  # no item at key


    def remIoVals(self, sdb, key, *, sep=b'.'):
        """Deletes all values at apparent effective key.
        Uses hidden ordinal key suffix for insertion ordering.
        The suffix is appended and stripped transparently.

        Returns:
            result (bool): True if at least one value was deleted at key.
                           False otherwise, if no values at key

        Parameters:
            db (lmdb._Database): instance of named sub db with dupsort==False
            key (bytes): Apparent effective key
        """
        result = False
        with self.env.begin(db=sdb, write=True, buffers=True) as txn:
            iokey = self.suffix(key, 0, sep=sep)  # start at zeroth value for key
            cursor = txn.cursor()
            if cursor.set_range(iokey):  # move to val at key >= iokey if any
                iokey = cursor.key()
                while iokey:  # end of database iokey == b'' cant internext.
                    ckey, cion = self.unsuffix(iokey, sep=sep)
                    if ckey != key:  # past key
                        break
                    result = cursor.delete() or result  # delete moves cursor to next item
                    iokey = cursor.key()  # cursor now at next item after deleted
            return result


    def cntIoVals(self, sdb, key, *, sep=b'.'):
        """Count all values with the same apparent effective key.
        Uses hidden ordinal key suffix for insertion ordering.
        The suffix is appended and stripped transparently.

        Returns:
            count (int): count values in set at apparent effective key

        Parameters:
            db (lmdb._Database): instance of named sub db with dupsort==False
            key (bytes): Apparent effective key
        """
        return len(self.getIoVals(sdb=sdb, key=key, sep=sep))


    def getTopIoItemIter(self, sdb, top=b'', *, sep=b'.'):
        """Gets Iterator of all values in branch rooted at top key
        The suffix is appended and stripped transparently.

        Returns:
            items (Iterator[(key,val)]): iterator of tuples (key, val) where
                key is apparent key with hidden insertion ordering suffixe removed
                from effective key. Iterates over top branch of insertion
                ordered set values where each effective key has trailing hidden
                suffix of serialization of insertion ordering ordinal.

            Uses hidden ordinal key suffix for insertion ordering.

        Raises StopIteration Error when empty.

        Parameters:
            db (lmdb._Database): instance of named sub db with dupsort==False
            top (bytes): top key in db. When top is empty then every item in db.
            sep (bytes): sep character for attached io suffix
        """
        for iokey, val in self.getTopItemIter(sdb=sdb, top=top):
            key, ion = self.unsuffix(iokey, sep=sep)
            yield (key, val)


    # insertion ordered queue pure FIFO not set. No dedup.
    def addIoVal(self, sdb, key, val, *, sep=b'.'):
        """Add val at end (append) of insertion ordered set of values all with the
        same apparent effective key. Does not dedup val.
        Uses hidden ordinal key suffix for insertion ordering.
        The suffix is appended and stripped transparently.

        Returns:
           result (bool): True is added to set. False if already in set.

        Parameters:
            db (lmdb._Database): instance of named sub db with dupsort==False
            key (bytes): Apparent effective key
            val (bytes): serialized value to add

        """
        with self.env.begin(db=sdb, write=True, buffers=True) as txn:
            ion = 0
            iokey = self.suffix(key, ion, sep=sep)  # start zeroth entry if any
            cursor = txn.cursor()
            if cursor.set_range(iokey):  # move to val at key >= iokey if any
                for iokey in cursor.iternext(values=False):  # get iokey at cursor
                    ckey, cion = self.unsuffix(iokey, sep=sep)
                    if ckey == key:  # found one
                        ion = cion + 1  # next ion is increment of found cion
                    else:  # prev entry if any was the last entry for key
                        break  # done

            iokey = self.suffix(key, ion, sep=sep)  # ion is 1 after last
            return cursor.put(iokey, val, dupdata=False, overwrite=True)


    def putIoVals(self, sdb, key, vals, *, sep=b'.'):
        """Adds at end (append) each val in vals to insertion ordered list of
        values all with the same apparent effective key for each val.
        Does not dedup added vals.
        Uses hidden ordinal key suffix for insertion ordering.
        The suffix is appended and stripped transparently.

        Returns:
           result (bool): True if added to set. False if already in set.

        Parameters:
            db (lmdb._Database): instance of named sub db with dupsort==False
            key (bytes): Apparent effective key
            vals (Iterable): serialized values to add to set of vals at key

        """
        result = False
        with self.env.begin(db=sdb, write=True, buffers=True) as txn:
            ion = 0  # start at zeroth one
            iokey = self.suffix(key, ion, sep=sep)  # start zeroth entry if any
            cursor = txn.cursor()
            if cursor.set_range(iokey):  # move to val at key >= iokey if any
                for iokey in cursor.iternext(values=False):  # get iokey at cursor
                    ckey, cion = self.unsuffix(iokey, sep=sep)
                    if ckey == key:  # found one
                        ion = cion + 1  # next ion is increment of found cion
                    else:
                        break  # prev entry if any was the last entry for key


            for i, val in enumerate(vals):  # i is zero based
                iokey = self.suffix(key, ion+i, sep=sep)  # ion is 1 after last
                result = cursor.put(iokey,
                                    val,
                                    dupdata=False,
                                    overwrite=True)
            return result


    def pinIoVals(self, sdb, key, vals, *, sep=b'.'):
        """Erase all vals at key and then add vals as insertion ordered set of
        values all with the same apparent effective key. Does not dedup vals.
        Uses hidden ordinal key suffix for insertion ordering.
        The suffix is appended and stripped transparently.

        Returns:
           result (bool): True is added to set.

        Parameters:
            db (lmdb._Database): instance of named sub db with dupsort==False
            key (bytes): Apparent effective key
            vals (abc.Iterable): serialized values to add to set of vals at key
        """
        self.remIoVals(sdb=sdb, key=key, sep=sep)
        result = False
        with self.env.begin(db=sdb, write=True, buffers=True) as txn:
            for i, val in enumerate(vals):  # starts at zero
                iokey = self.suffix(key, i, sep=sep)  # ion is at add on amount
                result = txn.put(iokey, val, dupdata=False, overwrite=True)
            return result


    # IoSet insertion ordered set with FIFO access queue. Set dedups entries
    def addIoSetVal(self, sdb, key, val, *, sep=b'.'):
        """Add val idempotently to insertion ordered set of values all with the
        same apparent effective key if val not already in set of vals at key.
        Dedups the added val.
        Uses hidden ordinal key suffix for insertion ordering.
        The suffix is appended and stripped transparently.

        Returns:
           result (bool): True is added to set. False if already in set.

        Parameters:
            db (lmdb._Database): instance of named sub db with dupsort==False
            key (bytes): Apparent effective key
            val (bytes): serialized value to add

        """
        with self.env.begin(db=sdb, write=True, buffers=True) as txn:
            vals = oset()
            ion = 0
            iokey = self.suffix(key, ion, sep=sep)  # start zeroth entry if any
            cursor = txn.cursor()
            if cursor.set_range(iokey):  # move to val at key >= iokey if any
                for iokey, cval in cursor.iternext():  # get iokey, val at cursor
                    ckey, cion = self.unsuffix(iokey, sep=sep)
                    if ckey == key:
                        vals.add(cval)  # another entry at key
                        ion = cion + 1 # ion to add at is increment of cion
                    else:  # prev entry if any was the last entry for key
                        break  # done

            if val in vals:  # already in set
                return False

            iokey = self.suffix(key, ion, sep=sep)  # ion is at add on amount
            return cursor.put(iokey, val, dupdata=False, overwrite=False)


    def putIoSetVals(self, sdb, key, vals, *, sep=b'.'):
        """Adds idempotently each val in vals to insertion ordered set of values
        all with the same apparent effective key for each val that is not already
        in set of vals at key. Dedups the put vals.
        Uses hidden ordinal key suffix for insertion ordering.
        The suffix is appended and stripped transparently.

        Returns:
           result (bool): True if added to set. False if already in set.

        Parameters:
            db (lmdb._Database): instance of named sub db with dupsort==False
            key (bytes): Apparent effective key
            vals (Iterable): serialized values to add to set of vals at key

        """
        result = False
        vals = oset(vals)  # make set
        with self.env.begin(db=sdb, write=True, buffers=True) as txn:
            ion = 0
            iokey = self.suffix(key, ion, sep=sep)  # start zeroth entry if any
            cursor = txn.cursor()
            if cursor.set_range(iokey):  # move to val at key >= iokey if any
                pvals = oset()  # pre-existing vals at key
                for iokey, cval in cursor.iternext():  # get iokey, val at cursor
                    ckey, cion = self.unsuffix(iokey, sep=sep)
                    if ckey == key:
                        pvals.add(cval)  # set add idempotently another entry at key
                        ion = cion + 1  # ion to add at is increment of cion
                    else:  # prev entry if any was the last entry for key
                        break  # done
                vals -= pvals  # set math remove vals already in pvals

            for i, val in enumerate(vals):
                iokey = self.suffix(key, ion+i, sep=sep)  # ion is at add on amount
                result = cursor.put(iokey,
                                    val,
                                    dupdata=False,
                                    overwrite=False) or result  # not short circuit
            return result


    def pinIoSetVals(self, sdb, key, vals, *, sep=b'.'):
        """Erase all vals at key and then add unique vals as insertion ordered set of
        values all with the same apparent effective key. Dedups the set vals.
        Uses hidden ordinal key suffix for insertion ordering.
        The suffix is appended and stripped transparently.

        Returns:
           result (bool): True is added to set.

        Parameters:
            db (lmdb._Database): instance of named sub db with dupsort==False
            key (bytes): Apparent effective key
            vals (abc.Iterable): serialized values to add to set of vals at key
        """
        self.remIoVals(sdb=sdb, key=key, sep=sep)
        result = False
        vals = oset(vals)  # make set
        with self.env.begin(db=sdb, write=True, buffers=True) as txn:
            for i, val in enumerate(vals):
                iokey = self.suffix(key, i, sep=sep)  # ion is at add on amount
                result = txn.put(iokey, val, dupdata=False, overwrite=True) or result
            return result


    def remIoSetVal(self, sdb, key, val, *, sep=b'.'):
        """Removes (delete) matching val at apparent effective key if exists.
        Uses hidden ordinal key suffix for insertion ordering.
        The suffix is appended and stripped transparently.

        Because the insertion order of val is not provided must perform a linear
        search over set of values.

        Another problem is that vals may get added and deleted in any order so
        the max suffix ion may creep up over time. The suffix ordinal max > 2**16
        is an impossibly large number, however, so the suffix will not max out
        practically.But its not the most elegant solution.

        In some cases a better approach would be to use getIoSetItemsIter which
        returns the actual iokey not the apparent effetive key so can delete
        using the iokey without searching for the value. This is most applicable
        when processing escrows where all the escrowed items are processed linearly
        and one needs to delete some of them in stride with their processing.

        Returns:
            result (bool): True if val was deleted at key. False otherwise
                if val not found at key

        Parameters:
            db (lmdb._Database): instance of named sub db with dupsort==False
            key (bytes): Apparent effective key
            val (bytes): value to delete
        """
        with self.env.begin(db=sdb, write=True, buffers=True) as txn:
            iokey = self.suffix(key, 0, sep=sep)  # start zeroth value for key
            cursor = txn.cursor()
            if cursor.set_range(iokey):  # move to val at key >= iokey if any
                for iokey, cval in cursor.iternext():  # get iokey, val at cursor
                    ckey, cion = self.unsuffix(iokey, sep=sep)
                    if ckey != key:  # prev entry if any was the last entry for key
                        break  # done
                    if val == cval:
                        return cursor.delete()  # delete also moves to next so doubly moved
            return False


@contextmanager
def openDuror(*, cls=None, name="test", temp=True, **kwa):
    """Context manager wrapper for Duror instances.
    Defaults to temporary databases.
    Context 'with' statements call .close on exit of 'with' block

    Parameters:
        cls (Type[Duror]): Class (subclass) defaults to Duror when None
        name (str):  of Duror dirPath so can have multiple holders
             that each use different dirPath name
        temp (bool):  True means open in temporary directory, clear on close
                    Otherwise open in persistent directory, do not clear on close

    Usage:

    with openHolder(name="gen1") as baser1:
        baser1.env  ....

    with openHolder(name="gen2, cls=Baser)

    """
    duror = None
    if cls is None:
        cls = Duror
    try:
        duror = cls(name=name, temp=temp, reopen=True, **kwa)
        yield duror

    finally:
        if duror:
            duror.close(clear=duror.temp)  # clears if lmdber.temp



class DurorDoer(Doer):
    """Duror Doer

    Attributes:
        done (bool): completion state:
            True means completed
            Otherwise incomplete. Incompletion maybe due to close or abort.
        duror (Duror): instance

    Inherited Properties:
        tyme (float): relative cycle time of associated Tymist .tyme obtained
            via injected .tymth function wrapper closure.
        tymth (func): closure returned by Tymist .tymeth() method.
            When .tymth is called it returns associated Tymist .tyme.
            .tymth provides injected dependency on Tymist tyme base.
        tock (float)): desired time in seconds between runs or until next run,
                 non negative, zero means run asap

    """

    def __init__(self, duror, **kwa):
        """Init instance

        Inherited Parameters:
           tymist (Tymist): instance
           tock (float): initial value of .tock in seconds

        Parameters:
           duror (Duror): instance
        """
        super(DurorDoer, self).__init__(**kwa)
        self.duror = duror


    def enter(self, *, temp=None):
        """Do 'enter' context actions. Override in subclass. Not a generator method.
        Set up resources. Comparable to context manager enter.

        Parameters:
            temp (bool | None): True means use temporary file resources if any
                                None means ignore parameter value. Use self.temp

        Inject temp or self.temp into file resources here if any
        """
        # inject temp into file resources here if any
        if not self.duror.opened:
            self.duror.reopen(temp=temp)


    def exit(self):
        """"""
        self.duror.close(clear=self.duror.temp)



class SuberBase():
    """Base class for Sub DBs of Duror
    Provides common methods for subclasses
    Do not instantiate but use a subclass

    Class Attribues:
        Sep (str): default separator to convert keys iterator to key bytes for db key

    Attributes:
        db (dbing.LMDBer): base LMDB db
        sdb (lmdb._Database): instance of lmdb named sub db for this Suber
        sep (str): separator for combining keys tuple of strs into key bytes
        verify (bool): True means reverify when ._des from db when applicable
                       False means do not reverify. Default False
    """
    Sep = '_'  # separator for combining key iterables

    def __init__(self, db: Duror, *,
                       subkey: str='docs.',
                       dupsort: bool=False,
                       sep: str=None,
                       verify: bool=False,
                       **kwa):
        """Parameters:
            db (dbing.LMDBer): base db
            subkey (str):  LMDB sub database key
            dupsort (bool): True means enable duplicates at each key
                               False (default) means do not enable duplicates at
                               each key
            sep (str): separator to convert keys iterator to key bytes for db key
                       default is self.Sep == '.'
            verify (bool): True means reverify when ._des from db when applicable
                           False means do not reverify. Default False
        """
        super(SuberBase, self).__init__()  # for multi inheritance
        self.db = db
        self.sdb = self.db.env.open_db(key=subkey.encode("utf-8"), dupsort=dupsort)
        self.sep = sep if sep is not None else self.Sep
        self.verify = True if verify else False



    def _tokey(self, keys: str|bytes|memoryview|Iterable[str|bytes|memoryview],
                topive: bool=False):
        """Converts keys to key bytes with proper separators and returns key bytes.
        If keys is already str or bytes or memoryview then returns key bytes.
        Else If keys is iterable (non-str) of strs or bytes then joins with
        separator converts to key bytes and returns. When keys is iterable and
        topive is True then enables partial key from top branch of key space given
        by partial keys by appending separator to end of partial key

        Returns:
           key (bytes): each element of keys is joined by .sep. If topive then
                        last char of key is .sep

        Parameters:
           keys (str | bytes | memoryview | Iterable[str | bytes]): db key or
                        Iterable of (str | bytes) to form key.
           topive (bool): True means treat as partial key tuple from top branch of
                       key space given by partial keys. Resultant key ends in .sep
                       character.
                       False means treat as full branch in key space. Resultant key
                       does not end in .sep character.
                       When last item in keys is empty str then will treat as
                       partial ending in sep regardless of topive value

        """
        if hasattr(keys, "encode"):  # str
            return keys.encode("utf-8")
        if isinstance(keys, memoryview):  # memoryview of bytes
            return bytes(keys)  # return bytes
        elif hasattr(keys, "decode"): # bytes
            return keys
        if topive and keys[-1]:  # topive and keys is not already partial
            keys = tuple(keys) + ('',)  # cat empty str so join adds trailing sep
        return (self.sep.join(key.decode() if hasattr(key, "decode") else key
                              for key in keys).encode())


    def _tokeys(self, key: str | bytes | memoryview):
        """Converts key bytes to keys tuple of strs by decoding and then splitting
        at separator .sep.

        Returns:
           keys (tuple[str]): makes tuple by splitting key at sep

        Parameters:
           key (str | bytes | memoryview): db key.

        """
        if isinstance(key, memoryview):  # memoryview of bytes
            key = bytes(key)
        if hasattr(key, "decode"):  # bytes
            key = key.decode()  # convert to str
        return tuple(key.split(self.sep))


    def _ser(self, val: str|bytes|memoryview):
        """
        Serialize value to bytes to store in db
        Parameters:
            val (str | bytes | memoryview): encodable as bytes
        """
        if isinstance(val, memoryview):  # memoryview is always bytes
            val = bytes(val)  # return bytes
        return (val.encode() if hasattr(val, "encode") else val)


    def _des(self, val: bytes|memoryview):
        """Deserialize val to str
        Parameters:
            val (bytes | memoryview): decodable as str
        """
        if isinstance(val, memoryview):  # memoryview is always bytes
            val = bytes(val)  # convert to bytes
        return (val.decode() if hasattr(val, "decode") else val)


    def trim(self, keys: str|bytes|memoryview|Iterable=b"", *, topive=False):
        """Removes all entries whose keys startswith keys. Enables removal of whole
        branches of db key space. To ensure that proper separation of a branch
        include empty string as last key in keys. For example ("a","") deletes
        'a.1'and 'a.2' but not 'ab'

        Parameters:
            keys (Iteratabke[str | bytes | memoryview]): of key parts that may be
                a truncation of a full keys tuple in  in order to address all the
                items from multiple branches of the key space.
                If keys is empty then trims all items in database.
                Either append "" to end of keys Iterable to ensure get properly
                separated top branch key or use top=True.

            topive (bool): True means treat as partial key tuple from top branch of
                       key space given by partial keys. Resultant key ends in .sep
                       character.
                       False means treat as full branch in key space. Resultant key
                       does not end in .sep character.
                       When last item in keys is empty str then will treat as
                       partial ending in sep regardless of top value


        Returns:
           result (bool): True if val at key exists so delete successful. False otherwise
        """
        return(self.db.remTopVals(sdb=self.sdb, top=self._tokey(keys, topive=topive)))


    def getFullItemIter(self, keys: str|bytes|memoryview|Iterable[str|bytes]="",
                       *, topive=False):
        """Iterator over items in .db that returns full items with subclass
        specific special hidden parts shown for debugging or testing.

        Returns:
            items (Iterator[tuple[key,val]]): (key, val) tuples of each item
            over the all the items in subdb whose key startswith key made from
            keys. Keys may be keyspace prefix to return branches of key space.
            When keys is empty then returns all items in subdb.
            This is meant to return full parts of items in both keyspace and
            valuespace which may be useful in debugging or testing.

        Parameters:
            keys (str|bytes|memoryview|Iteratable[str | bytes | memoryview]):
                of key parts that may be
                a truncation of a full keys tuple in  in order to address all the
                items from multiple branches of the key space.
                If keys is empty then gets all items in database.
                Either append "" to end of keys Iterable to ensure get properly
                separated top branch key or use top=True.
                In Python str.startswith('') always returns True so if branch
                key is empty string it matches all keys in db with startswith.


            topive (bool): True means treat as partial key tuple from top branch of
                       key space given by partial keys. Resultant key ends in .sep
                       character.
                       False means treat as full branch in key space. Resultant key
                       does not end in .sep character.
                       When last item in keys is empty str then will treat as
                       partial ending in sep regardless of top value

        """
        for key, val in self.db.getTopItemIter(sdb=self.sdb,
                                               top=self._tokey(keys, topive=topive)):
            yield (self._tokeys(key), self._des(val))


    def getItemIter(self, keys: str|bytes|memoryview|Iterable="",
                       *, topive=False):
        """Iterator over items in .db subclasses that do special hidden transforms
        on either the keyspace or valuespace should override this method to hide
        hidden parts from the returned items. For example, adding either
        a hidden key space suffix or hidden val space proem to ensure insertion
        order. Use getFullItemIter instead to return full items with hidden parts
        shown for debugging or testing.

        Returns:
            items (Iterator[tuple[key,val]]): (key, val) tuples of each item
            over the all the items in subdb whose key startswith key made from
            keys. Keys may be keyspace prefix to return branches of key space.
            When keys is empty then returns all items in subdb



        Parameters:
            keys (str|bytes|memoryview|Iterable[str|bytes|memoryview]): of key
                parts that may be
                a truncation of a full keys tuple in  in order to address all the
                items from multiple branches of the key space.
                If keys is empty then gets all items in database.
                Either append "" to end of keys Iterable to ensure get properly
                separated top branch key or use top=True.
                In Python str.startswith('') always returns True so if branch
                key is empty string it matches all keys in db with startswith.


            topive (bool): True means treat as partial key tuple from top branch of
                key space given by partial keys. Resultant key ends in .sep
                character.
                False means treat as full branch in key space. Resultant key
                does not end in .sep character.
                When last item in keys is empty str then will treat as
                partial ending in sep regardless of top value

        """
        for key, val in self.db.getTopItemIter(sdb=self.sdb,
                                               top=self._tokey(keys, topive=topive)):
            yield (self._tokeys(key), self._des(val))

    def cntAll(self):
        """Counts all entries in subdb self.sdb

        Returns:
            count (int): all entries in .sdb
        """
        return self.db.cntVals(sdb=self.sdb)



class Suber(SuberBase):
    """Subclass of SuberBase with no LMDB duplicates (i.e. multiple values at same key).

    Inherited Class Attribues:
        Sep (str): default separator to convert keys iterator to key bytes for db key

    Inherited Attributes:
        db (dbing.LMDBer): base LMDB db
        sdb (lmdb._Database): instance of lmdb named sub db for this Suber
        sep (str): separator for combining keys tuple of strs into key bytes
        verify (bool): True means reverify when ._des from db when applicable
                       False means do not reverify. Default False

    """

    def __init__(self, db: Duror, *,
                       subkey: str = 'docs.',
                       dupsort: bool=False, **kwa):
        """
        Inherited Parameters:
            db (dbing.LMDBer): base db
            subkey (str):  LMDB sub database key
            dupsort (bool): True means enable duplicates at each key
                               False (default) means do not enable duplicates at
                               each key. Set to False
            sep (str): separator to convert keys iterator to key bytes for db key
                       default is self.Sep == '.'
            verify (bool): True means reverify when ._des from db when applicable
                           False means do not reverify. Default False

        Parameters:
            db (dbing.LMDBer): base db
            subkey (str):  LMDB sub database key
        """
        super(Suber, self).__init__(db=db, subkey=subkey, dupsort=False, **kwa)


    def put(self, keys: str|bytes|memoryview|Iterable[str|bytes|memoryview],
                  val: bytes|str|memoryview):
        """
        Puts val at key made from keys. Does not overwrite

        Parameters:
            keys (str|bytes|memoryview|Iterable[str|bytes|memoryview]): of key
                                      strs to be combined in order to form key
            val (bytes|str|memoryview): value from ._ser

        Returns:
            result (bool): True If successful, False otherwise, such as key
                              already in database.
        """
        return (self.db.putVal(sdb=self.sdb,
                               key=self._tokey(keys),
                               val=self._ser(val)))


    def pin(self, keys: str|bytes|memoryview|Iterable[str|bytes|memoryview],
                  val: bytes|str|memoryview):
        """
        Pins (sets) val at key made from keys. Overwrites.

        Parameters:
            keys (str|bytes|memoryview|Iterable[str|bytes|memoryview]): of key
                                      strs to be combined in order to form key
            val (bytes|str|memoryview): value from ._ser

        Returns:
            result (bool): True If successful. False otherwise.
        """
        return (self.db.pinVal(sdb=self.sdb,
                               key=self._tokey(keys),
                               val=self._ser(val)))


    def get(self, keys:  str|bytes|memoryview|Iterable[str|bytes|memoryview]):
        """
        Gets val at keys

        Parameters:
            keys ( str|bytes|memoryview|Iterable[str|bytes|memoryview]): of key
                                       strs to be combined in order to form key

        Returns:
            data (str): decoded as utf-8 or whatever ._des provides
            None if no entry at keys

        Usage:
            Use walrus operator to catch and raise missing entry
            if (data := mydb.get(keys)) is None:
                raise ExceptionHere
            use data here

        """
        val = self.db.getVal(sdb=self.sdb, key=self._tokey(keys))
        return (self._des(val) if val is not None else None)


    def rem(self, keys:  str|bytes|memoryview|Iterable[str|bytes|memoryview]):
        """
        Removes entry at keys

        Parameters:
            keys ( str|bytes|memoryview|Iterable[str|bytes|memoryview]): of key
                                       strs to be combined in order to form key

        Returns:
           result (bool): True if key exists so delete successful. False otherwise
        """
        return(self.db.remVal(sdb=self.sdb, key=self._tokey(keys)))



class IoSuber(SuberBase):
    """Insertion Ordered Suber class that supports a durable queue (list) of
    entries at a given effective database key but with dupsort==False. Does not
    dedup entries. For dedup see IoSetSuber.

    Effective data model is that there are multiple values a list of values
    where every entry has the same key (duplicate key).
    The list is ordered using insertion order. Any given value may appear
    multiple times in the list.

    Here the key is augmented with a hidden numbered suffix that provides a
    an ordinal that makes each key unique per entry.  The effective key includes
    the ordinal suffix. The suffix is appended and stripped transparently. The
    list of multiple items can be retrieved in insertion order when iterating
    or as a list of elements or as FIFO queue.

    Inherited Class Attribues:
        Sep (str): default separator to convert keys iterator to key bytes for db key

    ClassAttributes:
        IonSep (str): default separator to suffix insertion order ordinal number

    Inherited Attributes:
        db (dbing.LMDBer): base LMDB db
        sdb (lmdb._Database): instance of lmdb named sub db for this Suber
        sep (str): separator for combining keys tuple of strs into key bytes
        verify (bool): True means reverify when ._des from db when applicable
                       False means do not reverify. Default False

    Attributes
        ionsep (str): separator to suffix insertion order ordinal number
                       default is self.IonSep == '.'
    """
    IonSep = '.'  # separator for suffixing insertion order ordinal number

    def __init__(self, db: dbing.LMDBer, *,
                       subkey: str='docs.',
                       dupsort: bool=False,
                       ionsep: str=None, **kwa):
        """Initialize instance

        Inherited Parameters:
            db (dbing.LMDBer): base db
            subkey (str):  LMDB sub database key
            dupsort (bool): True means enable duplicates at each key
                               False (default) means do not enable duplicates at
                               each key
            sep (str): separator to convert keys iterator to key bytes for db key
                       default is self.Sep == '.'
            verify (bool): True means reverify when ._des from db when applicable
                           False means do not reverify. Default False

        Parameters:
            ionsep (str|None): separator to suffix insertion order ordinal number
                       default is self.IonSep == '.'
        """
        super(IoSuber, self).__init__(db=db, subkey=subkey, dupsort=False, **kwa)
        self.ionsep = ionsep if ionsep is not None else self.IonSep


    def getFirst(self, keys: str | bytes | memoryview | Iterable):
        """Gets first val inserted at effecive key made from keys and hidden ordinal
        suffix.

        Parameters:
            keys (Iterable): of key strs to be combined in order to form key

        Returns:
            val (str|None):  value str, None if no entry at keys

        """
        val = self.db.getIoValFirst(sdb=self.sdb, key=self._tokey(keys))
        return (self._des(val) if val is not None else val)


    def getLast(self, keys: str | bytes | memoryview | Iterable):
        """Gets last val inserted at effecive key made from keys and hidden ordinal
        suffix.

        Parameters:
            keys (Iterable): of key strs to be combined in order to form key

        Returns:
            val (str|None):  value str, None if no entry at keys

        """
        val = self.db.getIoValLast(sdb=self.sdb, key=self._tokey(keys))
        return (self._des(val) if val is not None else val)


    def get(self, keys: str | bytes | memoryview | Iterable):
        """Gets vals set list at key made from effective keys

        Parameters:
            keys (Iterable): of key strs to be combined in order to form key

        Returns:
            vals (Iterable):  each item in list is str
                          empty list if no entry at keys

        """
        return ([self._des(val) for val in
                    self.db.getIoValsIter(sdb=self.sdb,
                                             key=self._tokey(keys),
                                             sep=self.ionsep)])


    def getIter(self, keys: str | bytes | memoryview | Iterable):
        """Gets vals iterator at effecive key made from keys and hidden ordinal suffix.
        All vals in set of vals that share same effecive key are retrieved in
        insertion order.

        Parameters:
            keys (Iterable): of key strs to be combined in order to form key

        Returns:
            vals (Iterator):  str values. Raises StopIteration when done

        """
        for val in self.db.getIoValsIter(sdb=self.sdb,
                                            key=self._tokey(keys),
                                            sep=self.ionsep):
            yield self._des(val)


    def pop(self, keys: str | bytes | memoryview | Iterable):
        """Pops first val if any inserted at effecive key made from keys and
        hidden ordinal suffix. Pop returns and deletes value if any.

        Parameters:
            keys (Iterable): of key strs to be combined in order to form key

        Returns:
            val (str|None):  value str, None if no entry at keys

        """
        val = self.db.popIoVal(sdb=self.sdb, key=self._tokey(keys))
        return (self._des(val) if val is not None else val)


    def add(self, keys: str | bytes | memoryview | Iterable,
            val: str | bytes | memoryview):
        """Add val idempotently to vals at effective key made from keys and hidden
        ordinal suffix. Idempotent means that added value is not already in set
        of vals at key. Does not overwrite or add same value at same key more
        than once.

        Parameters:
            keys (str | bytes | memoryview | Iterable): of key parts to be
                    combined in order to form key
            val (str | bytes | memoryview): serialization

        Returns:
            result (bool): True means value added ,
                            False otherwise.

        """
        return (self.db.addIoVal(sdb=self.sdb,
                                    key=self._tokey(keys),
                                    val=self._ser(val),
                                    sep=self.ionsep))


    def put(self, keys: str | bytes | memoryview | Iterable,
                  vals: str | bytes | memoryview | Iterable):
        """Puts all vals each with an effective key made from keys and hidden
        ordinal suffix that are not already in set of vals at key.
        Does not overwrite existing vals. Does not add val if already same val
        in set.

        Parameters:
            keys (str | bytes | memoryview | Iterable): of key parts to be
                    combined in order to form key
            vals (str | bytes | memoryview | Iterable): of str serializations

        Returns:
            result (bool): True If successful, False otherwise.

        """
        if not isNonStringIterable(vals):  # not iterable
            vals = (vals, )  # make iterable
        return (self.db.putIoVals(sdb=self.sdb,
                                     key=self._tokey(keys),
                                     vals=[self._ser(val) for val in vals],
                                     sep=self.ionsep))


    def pin(self, keys: str | bytes | memoryview | Iterable,
                  vals: str | bytes | memoryview | Iterable):
        """Pins (sets) vals at effective key made from keys and hidden ordinal suffix.
        Overwrites. Removes all pre-existing vals that share same effective keys
        and replaces them with vals

        Parameters:
            keys (Iterable): of key strs to be combined in order to form key
            vals (Iterable): str serializations

        Returns:
            result (bool): True If successful, False otherwise.

        """
        if not isNonStringIterable(vals):  # not iterable
            vals = (vals, )  # make iterable
        return (self.db.pinIoVals(sdb=self.sdb,
                                     key=self._tokey(keys),
                                     vals=[self._ser(val) for val in vals],
                                     sep=self.ionsep))


    def rem(self, keys: str | bytes | memoryview | Iterable):
        """Removes entry at effective key made from keys and hidden ordinal suffix
        that matches val if any. Otherwise deletes all values at effective key.

        Parameters:
            keys (Iterable): of key strs to be combined in order to form key
            val (str):  value at key to delete. Subclass ._ser method may
                        accept different value types
                        if val is empty then remove all values at key

        Returns:
           result (bool): True if effective key with val exists so rem successful.
                           False otherwise

        """
        return self.db.remIoVals(sdb=self.sdb,
                                 key=self._tokey(keys),
                                 sep=self.ionsep)


    def cnt(self, keys: str | bytes | memoryview | Iterable):
        """Return count of  values at effective key made from keys and hidden ordinal
        suffix. Zero otherwise

        Parameters:
            keys (Iterable): of key strs to be combined in order to form key
        """
        return (self.db.cntIoVals(sdb=self.sdb,
                                     key=self._tokey(keys),
                                     sep=self.ionsep))


    def getItemIter(self, keys: str | bytes | memoryview | Iterable = "",
                    *, topive=False):
        """Return iterator over all the items in top branch defined by keys where
        keys may be truncation of full branch.

        Returns:
            items (Iterator): of (key, val) tuples over the all the items in
            subdb whose effective key startswith key made from keys.
            Keys may be keyspace prefix in order to return branches of key space.
            When keys is empty then returns all items in subdb.
            Returned key in each item has ordinal suffix removed.

        Parameters:
            keys (Iterable): tuple of bytes or strs that may be a truncation of
                a full keys tuple in  in order to address all the items from
                multiple branches of the key space. If keys is empty then gets
                all items in database.
                Either append "" to end of keys Iterable to ensure get properly
                separated top branch key or use top=True.
                In Python str.startswith('') always returns True so if branch
                key is empty string it matches all keys in db with startswith.

            topive (bool): True means treat as partial key tuple from top branch of
                key space given by partial keys. Resultant key ends in .sep
                character.
                False means treat as full branch in key space. Resultant key
                does not end in .sep character.
                When last item in keys is empty str then will treat as
                partial ending in sep regardless of top value

        """
        for key, val in self.db.getTopIoItemIter(sdb=self.sdb,
                top=self._tokey(keys, topive=topive), sep=self.ionsep):
            yield (self._tokeys(key), self._des(val))



class IoSetSuber(SuberBase):
    """Insertion Ordered Set Suber factory class that supports
    a set of distinct entries at a given effective database key but with
    dupsort==False. Effective data model is that there are multiple values in a
    set of values where every member of the set has the same key (duplicate key).
    The set of values is an ordered set using insertion order. Any given value
    may appear only once in the set (not a list).

    This works similarly to the IO value duplicates for the LMDBer class with a
    sub db  of LMDB (dupsort==True) but without its size limitation of 511 bytes
    for each value when dupsort==True.
    Here the key is augmented with a hidden numbered suffix that provides a
    an ordered set of values at each effective key (duplicate key). The suffix
    is appended and stripped transparently. The set of multiple items with
    duplicate keys are retrieved in insertion order when iterating or as a list
    of the set elements.

    Inherited Class Attribues:
        Sep (str): default separator to convert keys iterator to key bytes for db key

    ClassAttributes:
        IonSep (str): default separator to suffix insertion order ordinal number

    Inherited Attributes:
        db (dbing.LMDBer): base LMDB db
        sdb (lmdb._Database): instance of lmdb named sub db for this Suber
        sep (str): separator for combining keys tuple of strs into key bytes
        verify (bool): True means reverify when ._des from db when applicable
                       False means do not reverify. Default False

    Attributes
        ionsep (str): separator to suffix insertion order ordinal number
                       default is self.IonSep == '.'
    """
    IonSep = '.'  # separator for suffixing insertion order ordinal number

    def __init__(self, db: dbing.LMDBer, *,
                       subkey: str='docs.',
                       dupsort: bool=False,
                       ionsep: str=None, **kwa):
        """Initialize instance

        Inherited Parameters:
            db (dbing.LMDBer): base db
            subkey (str):  LMDB sub database key
            dupsort (bool): True means enable duplicates at each key
                               False (default) means do not enable duplicates at
                               each key
            sep (str): separator to convert keys iterator to key bytes for db key
                       default is self.Sep == '.'
            verify (bool): True means reverify when ._des from db when applicable
                           False means do not reverify. Default False

        Parameters:
            ionsep (str|None): separator to suffix insertion order ordinal number
                       default is self.IonSep == '.'
        """
        super(IoSetSuber, self).__init__(db=db, subkey=subkey, dupsort=False, **kwa)
        self.ionsep = ionsep if ionsep is not None else self.IonSep


    def getFirst(self, keys: str | bytes | memoryview | Iterable):
        """Gets first val inserted at effecive key made from keys and hidden ordinal
        suffix.

        Parameters:
            keys (Iterable): of key strs to be combined in order to form key

        Returns:
            val (str|None):  value str, None if no entry at keys

        """
        val = self.db.getIoValFirst(sdb=self.sdb, key=self._tokey(keys))
        return (self._des(val) if val is not None else val)


    def getLast(self, keys: str | bytes | memoryview | Iterable):
        """Gets last val inserted at effecive key made from keys and hidden ordinal
        suffix.

        Parameters:
            keys (Iterable): of key strs to be combined in order to form key

        Returns:
            val (str|None):  value str, None if no entry at keys

        """
        val = self.db.getIoValLast(sdb=self.sdb, key=self._tokey(keys))
        return (self._des(val) if val is not None else val)


    def get(self, keys: str | bytes | memoryview | Iterable):
        """Gets vals set list at key made from effective keys

        Parameters:
            keys (Iterable): of key strs to be combined in order to form key

        Returns:
            vals (Iterable):  each item in list is str
                          empty list if no entry at keys

        """
        return ([self._des(val) for val in
                    self.db.getIoValsIter(sdb=self.sdb,
                                             key=self._tokey(keys),
                                             sep=self.ionsep)])


    def getIter(self, keys: str | bytes | memoryview | Iterable):
        """Gets vals iterator at effecive key made from keys and hidden ordinal suffix.
        All vals in set of vals that share same effecive key are retrieved in
        insertion order.

        Parameters:
            keys (Iterable): of key strs to be combined in order to form key

        Returns:
            vals (Iterator):  str values. Raises StopIteration when done

        """
        for val in self.db.getIoValsIter(sdb=self.sdb,
                                            key=self._tokey(keys),
                                            sep=self.ionsep):
            yield self._des(val)


    def pop(self, keys: str | bytes | memoryview | Iterable):
        """Pops first val if any inserted at effecive key made from keys and
        hidden ordinal suffix. Pop returns and deletes value if any.

        Parameters:
            keys (Iterable): of key strs to be combined in order to form key

        Returns:
            val (str|None):  value str, None if no entry at keys

        """
        val = self.db.popIoVal(sdb=self.sdb, key=self._tokey(keys))
        return (self._des(val) if val is not None else val)


    def add(self, keys: str | bytes | memoryview | Iterable,
            val: str | bytes | memoryview):
        """Add val idempotently to vals at effective key made from keys and hidden
        ordinal suffix. Idempotent means that added value is not already in set
        of vals at key. Does not overwrite or add same value at same key more
        than once.

        Parameters:
            keys (str | bytes | memoryview | Iterable): of key parts to be
                    combined in order to form key
            val (str | bytes | memoryview): serialization

        Returns:
            result (bool): True means unique value added among duplications,
                            False means duplicate of same value already exists.

        """
        return (self.db.addIoSetVal(sdb=self.sdb,
                                    key=self._tokey(keys),
                                    val=self._ser(val),
                                    sep=self.ionsep))


    def put(self, keys: str | bytes | memoryview | Iterable,
                  vals: str | bytes | memoryview | Iterable):
        """Puts all vals each with an effective key made from keys and hidden
        ordinal suffix that are not already in set of vals at key.
        Does not overwrite existing vals. Does not add val if already same val
        in set.

        Parameters:
            keys (str | bytes | memoryview | Iterable): of key parts to be
                    combined in order to form key
            vals (str | bytes | memoryview | Iterable): of str serializations

        Returns:
            result (bool): True If successful, False otherwise.

        """
        if not isNonStringIterable(vals):  # not iterable
            vals = (vals, )  # make iterable
        return (self.db.putIoSetVals(sdb=self.sdb,
                                     key=self._tokey(keys),
                                     vals=[self._ser(val) for val in vals],
                                     sep=self.ionsep))


    def pin(self, keys: str | bytes | memoryview | Iterable,
                  vals: str | bytes | memoryview | Iterable):
        """Pins (sets) vals at effective key made from keys and hidden ordinal suffix.
        Overwrites. Removes all pre-existing vals that share same effective keys
        and replaces them with vals

        Parameters:
            keys (Iterable): of key strs to be combined in order to form key
            vals (Iterable): str serializations

        Returns:
            result (bool): True If successful, False otherwise.

        """
        if not isNonStringIterable(vals):  # not iterable
            vals = (vals, )  # make iterable
        return (self.db.pinIoSetVals(sdb=self.sdb,
                                     key=self._tokey(keys),
                                     vals=[self._ser(val) for val in vals],
                                     sep=self.ionsep))


    def rem(self, keys: str | bytes | memoryview | Iterable,
                   val: str | bytes | memoryview = b''):
        """Removes entry at effective key made from keys and hidden ordinal suffix
        that matches val if any. Otherwise deletes all values at effective key.

        Parameters:
            keys (Iterable): of key strs to be combined in order to form key
            val (str):  value at key to delete. Subclass ._ser method may
                        accept different value types
                        if val is empty then remove all values at key

        Returns:
            When val:
                result (bool): True if effective key with val exists so rem successful.
                           False otherwise
            Otherwise:
                result (bool): True if effective key with val exists so rem successful.
                           False otherwise

        """
        if val:
            return self.db.remIoSetVal(sdb=self.sdb,
                                       key=self._tokey(keys),
                                       val=self._ser(val),
                                       sep=self.ionsep)
        else:
            return self.db.remIoVals(sdb=self.sdb,
                                       key=self._tokey(keys),
                                       sep=self.ionsep)


    def cnt(self, keys: str | bytes | memoryview | Iterable):
        """Return count of  values at effective key made from keys and hidden ordinal
        suffix. Zero otherwise

        Parameters:
            keys (Iterable): of key strs to be combined in order to form key
        """
        return (self.db.cntIoVals(sdb=self.sdb,
                                     key=self._tokey(keys),
                                     sep=self.ionsep))


    def getItemIter(self, keys: str | bytes | memoryview | Iterable = "",
                    *, topive=False):
        """Return iterator over all the items in top branch defined by keys where
        keys may be truncation of full branch.

        Returns:
            items (Iterator): of (key, val) tuples over the all the items in
            subdb whose effective key startswith key made from keys.
            Keys may be keyspace prefix in order to return branches of key space.
            When keys is empty then returns all items in subdb.
            Returned key in each item has ordinal suffix removed.

        Parameters:
            keys (Iterable): tuple of bytes or strs that may be a truncation of
                a full keys tuple in  in order to address all the items from
                multiple branches of the key space. If keys is empty then gets
                all items in database.
                Either append "" to end of keys Iterable to ensure get properly
                separated top branch key or use top=True.
                In Python str.startswith('') always returns True so if branch
                key is empty string it matches all keys in db with startswith.

            topive (bool): True means treat as partial key tuple from top branch of
                key space given by partial keys. Resultant key ends in .sep
                character.
                False means treat as full branch in key space. Resultant key
                does not end in .sep character.
                When last item in keys is empty str then will treat as
                partial ending in sep regardless of top value

        """
        for key, val in self.db.getTopIoItemIter(sdb=self.sdb,
                top=self._tokey(keys, topive=topive), sep=self.ionsep):
            yield (self._tokeys(key), self._des(val))


class DomSuberBase(SuberBase):
    """Subclass of Suber with values that are serialized RegDom subclasses.

    forces .sep to '_'
    changes ._ser and ._des methods to serialize RegDom subclasses for db
    with default prosep for class name proem from json body is `\n`

    Prepends and strips RegDom subclass name as proem to value when serializing
    which is then stripped off when deserializing. Uses proem to lookup in
    RegDom._registry the class object to use to reinstantiate when deserializing.

    Inherited Class Attribues:
        Sep (str): default separator to convert keys iterator to key bytes for db key

    Inherited Attributes:
        db (dbing.LMDBer): base LMDB db
        sdb (lmdb._Database): instance of lmdb named sub db for this Suber
        sep (str): separator for combining keys tuple of strs into key bytes
        verify (bool): True means reverify when ._des from db when applicable
                       False means do not reverify. Default False

    ClassAttributes:
        ProSep (str): separator class name proem to serialized RegDom instance

    Attributes
        prosep (str): separator class name proem to serialized RegDom instance
                      default is self.ProSep == '\n'

    """
    ProSep = '\n'  # separator class name proem to serialized RegDom instance

    def __init__(self, db: Duror, *, subkey: str = 'bags.', sep='_', prosep=None,
                 **kwa):
        """
        Inherited Parameters:
            db (Duror): base db
            subkey (str):  LMDB sub database key
            dupsort (bool): True means enable duplicates at each key
                               False (default) means do not enable duplicates at
                               each key. Set to False
            sep (str): separator to convert keys iterator to key bytes for db key
                       default is self.Sep == '.'
            verify (bool): True means reverify when ._des from db when applicable
                           False means do not reverify. Default False

        Parameters:
            prosep (str|None): separator class name proem to serialized RegDom instance
                               default is self.ProSep == '\n'
        """
        super(DomSuberBase, self).__init__(db=db, subkey=subkey, sep=sep, **kwa)
        self.prosep = prosep if prosep is not None else self.ProSep


    def _ser(self, val: RegDom):
        """Serialize value to json bytes to store in db with proem that is
        class name. Must be in registry. Uses b'\n' as separator between
        proem class name and json serialization of val

        Parameters:
            val (RegDom): encodable as bytes json
        """
        try:
            klas = val._registry[val.__class__.__name__]
        except (KeyError, AttributeError) as ex:
            raise HierError(f"Unregistered subclass instance={val}") from ex

        return (klas.__name__.encode() + self.prosep.encode() + val._asjson())


    def _des(self, val: bytes|memoryview):
        """Deserialize val to RegDom subclass instance as given by proem
        prepended to json serialization of RegDom subclass instance
        Uses b'\n' as separator between proem class name and json serialization
        of val

        Parameters:
            val (bytes|memoryview): proem\njson

        Returns:
            dom (RegDom): instance of RegDom subclass as registered under proem

        """
        if isinstance(val, memoryview):  # memoryview is always bytes
            val = bytes(val)  # convert to bytes

        proem, ser = val.split(sep=self.prosep.encode(), maxsplit=1)
        proem = proem.decode()
        if proem in RegDom._registry:
            klas = RegDom._registry[proem]
        elif proem in IceRegDom._registry:
            klas = IceRegDom._registry[proem]
        else:
            raise HierError(f"Unregistered serialized subclass={proem}")

        return (klas._fromjson(ser))


class DomSuber(DomSuberBase, Suber):
    """Subclass of (DomSuberBase, Suber) with values that are serialized CanDom
    subclasses.

    forces .sep to '_'
    forces .prosep to '\n'

    Prepends and strips RegDom subclass name as proem to value when serializing
    which is then stripped off when deserializing. Uses proem to lookup in
    RegDom._registry the class object to use to reinstantiate when deserializing.

    Inherited Class Attribues:
        Sep (str): default separator to convert keys iterator to key bytes for db key
        ProSep (str): separator class name proem to serialized RegDom instance

    Inherited Attributes:
        db (dbing.LMDBer): base LMDB db
        sdb (lmdb._Database): instance of lmdb named sub db for this Suber
        sep (str): separator for combining keys tuple of strs into key bytes
        verify (bool): True means reverify when ._des from db when applicable
                       False means do not reverify. Default False
        prosep (str): separator class name proem to serialized RegDom instance
                      default is self.ProSep == '\n'

    """

    def __init__(self, db: Duror, *, subkey: str = 'cans.', sep='_', prosep='\n',
                 **kwa):
        """
        Inherited Parameters:
            db (Duror): base db
            subkey (str):  LMDB sub database key
            dupsort (bool): True means enable duplicates at each key
                               False (default) means do not enable duplicates at
                               each key. Set to False
            sep (str): separator to convert keys iterator to key bytes for db key
                       default is self.Sep == '.'
            verify (bool): True means reverify when ._des from db when applicable
                           False means do not reverify. Default False
            prosep (str|None): separator class name proem to serialized CanDom instance
                               default is self.ProSep == '\n'

        """
        super(DomSuber, self).__init__(db=db, subkey=subkey,
                                       sep=sep, prosep=prosep, **kwa)


class DomIoSuber(DomSuberBase, IoSuber):
    """Subclass of (DomSuberBase, IoSuber) with values that are serialized
    RegDom subclasses.  Insertion ordered list. No dedup.

    forces .ser to '_'
    forces .ionsep to '.'
    forces .prosep to '\n'

    Prepends and strips RegDom subclass name as proem to value when serializing
    which is then stripped off when deserializing. Uses proem to lookup in
    RegDom._registry the class object to use to reinstantiate when deserializing.

    Inherited Class Attribues:
        Sep (str): default separator to convert keys iterator to key bytes for db key
        ProSep (str): separator class name proem to serialized RegDom instance
        IonSep (str): default separator to suffix insertion order ordinal number

    Inherited Attributes:
        db (dbing.LMDBer): base LMDB db
        sdb (lmdb._Database): instance of lmdb named sub db for this Suber
        sep (str): separator for combining keys tuple of strs into key bytes
        verify (bool): True means reverify when ._des from db when applicable
                       False means do not reverify. Default False
        prosep (str): separator class name proem to serialized RegDom instance
                      default is self.ProSep == '\n'
        ionsep (str): separator to suffix insertion order ordinal number
                       default is self.IonSep == '.'

    """

    def __init__(self, db: Duror, *, subkey: str = 'dsqs.', sep='_',
                 prosep='\n', ionsep='.', **kwa):
        """
        Inherited Parameters:
            db (Duror): base db
            subkey (str):  LMDB sub database key
            dupsort (bool): True means enable duplicates at each key
                               False (default) means do not enable duplicates at
                               each key. Set to False
            sep (str): separator to convert keys iterator to key bytes for db key
                       default is self.Sep == '.'
            verify (bool): True means reverify when ._des from db when applicable
                           False means do not reverify. Default False
            prosep (str|None): separator class name proem to serialized CanDom instance
                               default is self.ProSep == '\n'
            ionsep (str|None): separator to suffix insertion order ordinal number
                       default is self.IonSep == '.'


        """
        super(DomIoSuber, self).__init__(db=db, subkey=subkey, sep=sep,
                                            prosep=prosep, ionsep=ionsep, **kwa)



class DomIoSetSuber(DomSuberBase, IoSetSuber):
    """Subclass of (DomSuberBase, IoSetSuber) with values that are serialized
    RegDom subclasses.  Insertion ordered set with dedup.

    forces .ser to '_'
    forces .ionsep to '.'
    forces .prosep to '\n'

    Prepends and strips RegDom subclass name as proem to value when serializing
    which is then stripped off when deserializing. Uses proem to lookup in
    RegDom._registry the class object to use to reinstantiate when deserializing.

    Inherited Class Attribues:
        Sep (str): default separator to convert keys iterator to key bytes for db key
        ProSep (str): separator class name proem to serialized RegDom instance
        IonSep (str): default separator to suffix insertion order ordinal number

    Inherited Attributes:
        db (dbing.LMDBer): base LMDB db
        sdb (lmdb._Database): instance of lmdb named sub db for this Suber
        sep (str): separator for combining keys tuple of strs into key bytes
        verify (bool): True means reverify when ._des from db when applicable
                       False means do not reverify. Default False
        prosep (str): separator class name proem to serialized RegDom instance
                      default is self.ProSep == '\n'
        ionsep (str): separator to suffix insertion order ordinal number
                       default is self.IonSep == '.'

    """

    def __init__(self, db: Duror, *, subkey: str = 'dsqs.', sep='_',
                 prosep='\n', ionsep='.', **kwa):
        """
        Inherited Parameters:
            db (Duror): base db
            subkey (str):  LMDB sub database key
            dupsort (bool): True means enable duplicates at each key
                               False (default) means do not enable duplicates at
                               each key. Set to False
            sep (str): separator to convert keys iterator to key bytes for db key
                       default is self.Sep == '.'
            verify (bool): True means reverify when ._des from db when applicable
                           False means do not reverify. Default False
            prosep (str|None): separator class name proem to serialized CanDom instance
                               default is self.ProSep == '\n'
            ionsep (str|None): separator to suffix insertion order ordinal number
                       default is self.IonSep == '.'


        """
        super(DomIoSetSuber, self).__init__(db=db, subkey=subkey, sep=sep,
                                            prosep=prosep, ionsep=ionsep, **kwa)



class Subery(Duror):
    """Subery subclass of Duror for managing subdbs of Duror for durable storage
    of action data
    """
    def __init__(self, **kwa):
        """
        Setup named sub databases.

        Inherited Parameters:  (see Duror)
        """
        super(Subery, self).__init__(**kwa)


    def reopen(self, **kwa):
        """Open sub databases

        Attributes:
            cans (Suber): subdb whose values are serialized Can instances
                Can is a durable Bag
            drqs (IoSetSub): subdb whose values are serialized RegDom instances
                Interfaced via a Durq which is a durable queue (FIFO)
            dsqs (IoSetSub): subdb whose values are serialized RegDom instances
                Interfaced via a Dusq which is  durable set queue (FIFO deduped)

        """
        super(Subery, self).reopen(**kwa)

        self.cans = DomSuber(db=self, subkey='cans.')
        self.drqs = DomIoSuber(db=self, subkey="drqs.")  # durable queue
        self.dsqs = DomIoSetSuber(db=self, subkey="dsqs.")  # durable set queue

        return self.env




# -*- encoding: utf-8 -*-
"""hio.base.hier.canning Module

Provides durable dom support for items in Hold

"""
from __future__ import annotations  # so type hints of classes get resolved later

from collections.abc import Mapping
from typing import Any, Type, ClassVar
from dataclasses import dataclass, astuple, asdict, field, fields, InitVar

from ...help import NonStringIterable, namify, registerify, TymeDom
from ...hioing import HierError
from ..during import DomSuber


@namify
@registerify
@dataclass
class CanDom(TymeDom):
    """CanDom is base class that adds support for durable storage via its
    ._sdb and ._key non-field attributes. Essentially acts as write through
    cache to durable copy in ._sdb at ._key when provided. Otherwise in memory
    only. Assignment to Hold instance at key, injects ._sdb and ._key. This
    makes the in-memory key in Hold and the durable key the same.

    Inherited Non-Field Class Attributes:
        _registry (ClassVar[dict]): dict of subclasses keyed by class.__name__
            Assigned by @registerify decorator
        _names (ClassVar[tuple[str]|None]): tuple of field names for class
            Assigned by @namify decorator

    Inherited Non-Field Attributes:
        _tymth (None|Callable): function wrapper closure returned by
            Tymist.tymen() method. When .tymth is called it returns associated
            Tymist.tyme. Provides injected dependency on Tymist cycle tyme base.
            None means not assigned yet.
            Use ._wind method to assign ._tymth after init of bag.
        _tyme (None|Float): cycle tyme of last update of a bag field.
            None means either ._tymth as not yet been assigned or this bag's
            fields have not yet been updated.

    Inherited Properties:
        _now (None|float): current tyme given by ._tymth if not None.

    Properties:
        _durable (bool): True means ._sdb and ._key and ._sdb.db and
                                .sdb.db.opened are not None
                         False otherwise

    Non-Field Attributes:
        _sdb (DomSuber|None): SuberBase subclass instance of durable subdb of Duror
        _key (str|None): database key used to store serialized field in ._cans
        _stale (bool): True means fields not yet been synced by write/read with durable
                       False means fields have been synced by write/read with durable
        _fresh (bool): True means fields being synced by read from durable
                       False means fields not being synced by read from durable
        _bulk (bool): True means do not write individual fields wait for bulk update
                      False means write individual fields as updated

    """
    _sdb: InitVar[None|DomSuber] = None  # durable storage of serialized fields
    _key: InitVar[None|str] = None  # durable storage of serialized fields
    _stale: InitVar[bool] = True  # fields synced by write to durable or not
    _fresh: InitVar[bool] = False  # fields synced by read from durable or not
    _bulk: InitVar[bool] = False  # bulk update or not


    def __post_init__(self, _tymth, _tyme, _sdb, _key, _stale, _fresh, _bulk):
        super(CanDom, self).__post_init__(_tymth, _tyme)
        self._sdb = _sdb
        self._key = _key
        self._stale = _stale
        self._fresh = _fresh
        self._bulk = _bulk


    #def __hash__(self):
        #"""Define hash so can work with ordered_set
        #__hash__ is not inheritable in dataclasses so must be explicitly defined
        #in every subclass
        #"""
        #return hash((self.__class__.__name__,) + self._astuple())  # almost same as __eq__


    def __setattr__(self, name, value):  # called by __setitem__
        super().__setattr__(name, value)
        if name in self._names:
            super().__setattr__("_tyme", self._now)
            if not self._bulk:
                self._pin()


    def _update(self, *pa, **kwa):
        """Use item update syntax
        """
        write = False
        try:
            self._bulk = True
            if pa:
                if len(pa) > 1:
                    raise TypeError(f"Expected 1 positional argument got {len(pa)}.")

                di = pa[0]
                if isinstance(di, Mapping):
                    for k, v in di.items():
                        self[k] = v
                        if k in self._names:
                            write = True
                elif isinstance(di, NonStringIterable):
                    for k, v in di:
                        self[k] = v
                        if k in self._names:
                            write = True
                else:
                    raise TypeError(f"Expected Mapping or NonStringIterable got"
                                    f" {type(di)}.")

            for k, v in kwa.items():
                self[k] = v
                if k in self._names:
                    write = True

            if write:
                self._pin()
        finally:
            self._bulk = False

    @property
    def _durable(self):
        """Property durable True when durable subdb injected and opened.

        Returns:
            durable (bool): True means ._sdb and ._key and ._sdb.db and
                                .sdb.db.opened are not None
                            False otherwise
        """
        return (self._sdb is not None and self._key is not None and self._sdb.db
                and self._sdb.db.opened)


    def _pin(self):
        """Writes own fields to ._sdb at ._key if any when not ._fresh.
        Sets ._stale to False on success
        """
        if self._durable and not self._fresh:
            self._sdb.pin(self._key, self)  # pin sdb._ser of own fields at key
            self._stale = False
            return True
        return False


    def _sync(self, force=False):
        """Syncs by checking exists durable copy in ._sdb at ._key. If so
        and ._stale or force then reads own fields from ._sdb at ._key if any.
        If not attempts to write to ._sdb at ._key
        Toggles ._fresh during read so can update own fields from read without
        triggering a write.

        Parameters:
            force (bool): True means force read even if not ._stale
                          Flase means do not force read
        """
        if self._durable and (self._stale or force):
            if can := self._sdb.get(self._key):  # not empty
                if can.__class__ != self.__class__:
                    raise HierError(f"Expected instance of "
                                    f"{self.__class__} on read, got "
                                    f"{can.__class__}")
                self._fresh = True  # so update does not write
                self._update(can._asdict())
                self._fresh = False
                self._stale = False  # now synced
                return True
            else:  # empty
                return self._pin()
        return False


@namify
@registerify
@dataclass
class Can(CanDom):
    """CanDom is base class that adds support for durable storage via its ._cans
    non-field attribute

    Inherited Non-Field Class Attributes:
        _registry (ClassVar[dict]): dict of subclasses keyed by class.__name__
            Assigned by @registerify decorator
        _names (ClassVar[tuple[str]|None]): tuple of field names for class
            Assigned by @namify decorator

    Inherited Non-Field Attributes:
        _tymth (None|Callable): function wrapper closure returned by
            Tymist.tymen() method. When .tymth is called it returns associated
            Tymist.tyme. Provides injected dependency on Tymist cycle tyme base.
            None means not assigned yet.
            Use ._wind method to assign ._tymth after init of bag.
        _tyme (None|Float): cycle tyme of last update of a bag field.
            None means either ._tymth as not yet been assigned or this bag's
            fields have not yet been updated.
        _sdb (DomSuber|None): SuberBase subclass instance of durable subdb of Duror
        _key (str|None): database key used to store serialized field in ._cans

    Inherited Properties:
        _now (None|float): current tyme given by ._tymth if not None.
        _durable (bool):  True means ._sdb and ._key are not None
                          False otherwise

    Field Attributes:
        value (Any):  generic value field
    """
    value: Any = None  # generic value


    def __hash__(self):
        """Define hash so can work with ordered_set
        __hash__ is not inheritable in dataclasses so must be explicitly defined
        in every subclass
        """
        return hash((self.__class__.__name__,) + self._astuple())  # almost same as __eq__




# -*- encoding: utf-8 -*-
"""
hio.base.hier.hiering Module

Provides hierarchical action support


Syntax notes for Mine and Dock use in boxwork:
    H hold (Hold) boxwork shared data attribute syntax with locally scoped
       variable H ref for hold .

    so need syntax does not heed to "quote" paths keys into the bags
    containers Mine dict subclasses with attribute support.

    H.root_dog.value  is equivalent to self.hold["root_dog"].value or
                                       self.hold[("root", "dog")].value

    So need term  "H.root_dog.value > 5" should compile directly and eval
    as long as H is in the locals() and H is a Mine instance.

    Likewise for
    D.root_dog.value where D is a Dock and root_dog is a key in the Dock.

    So no need to do substitutions or shorthand
    The hierarchy in the .mine/.dock is indicated by '_' separated keys
    The Box Boxer Actor names are forbidden from having '_" as an element
    with Renam regex test.



"""
from __future__ import annotations  # so type hints of classes get resolved later


from collections import deque, namedtuple
from collections.abc import Iterable, Mapping, Callable
from typing import Any, Type
from dataclasses import dataclass, astuple, asdict, field


from ..tyming import Tymee
from ... import hioing
from ...hioing import Mixin, HierError
from ...help import isNonStringIterable, MapDom, modify, Renam


"""Nabage (namedtuple):
Action nabes (contexts mileu neighborhood) for Acts

Fields:
   native (str): native nabe
   predo (str): predo nabe
   remark (str): remark sub-nabe of rendo
   rendo (str): rendo nabe
   enmark (str): enmark sub-nabe of endo
   endo (str): endo nabe
   redo (str): redo nabe
   afdo (str): afdo nabe
   godo (str): godo nabe
   exdo (str): exdo nabe
   rexdo (str): rexdo nabe
"""
Nabage = namedtuple("Nabage", "native predo remark rendo enmark endo"
                                      " redo afdo godo exdo rexdo")

Nabes = Nabage(native="native", predo="predo", remark="remark",
                     rendo="rendo", enmark="enmark", endo="endo",
                     redo="redo", afdo="afdo", godo="godo",
                     exdo="exdo", rexdo="rexdo")


@dataclass
class WorkDom(MapDom):
    """WorkDom provides state for building boxwork by a boxer to be injected
    make methods of Boxer by workify wrapper.


    Attributes:
        box (Box | None): current box in box work. None if not yet a box
        over (Box | None): current over Box in box work. None if top level
        bxpre (str):  default box name prefix used to generate unique box name
                    relative to boxer.boxes
        bxidx (int): default box name index used to generate unique box name
                    relative to boxer.boxes
        acts (dict):  registry of ActBase subclasses by name (including aliases)
        context (str): action context for act
    """
    box: None | Box = None  # current box in boxwork. None if not yet any box
    over: None | Box = None  # current over box in boxwork. None if not yet any over
    bxpre: str = 'box'  # default box name prefix when name not provided
    bxidx: int = 0  # default box name index when name not provided
    acts: dict = field(default_factory=dict)  # registry of Acts by name & aliases
    nabe: str = Nabes.native  # current action nabe (context)





# -*- encoding: utf-8 -*-
"""
hio.base.hier.hogging Module

Provides support for hold logging (hogging)


"""
from __future__ import annotations  # so type hints of classes get resolved later

import os
import uuid
from contextlib import contextmanager
import inspect
from collections import namedtuple, deque
from base64 import urlsafe_b64encode as encodeB64
from base64 import urlsafe_b64decode as decodeB64

from ...hioing import HierError
from ...help import timing  # import timing to pytest mock of nowIso8601 works
from ...help.helping import ocfn
from ..doing import Doer
from ..filing import Filer
from .hiering import Nabes
from .acting import ActBase, register


Ruleage = namedtuple("Rules", 'once every span update change')
Rules = Ruleage(once='once', every='every', span='span', update='update', change='change')

@register(names=('log', 'Log'))
class Hog(ActBase, Filer):
    """Hog is Act that supports metrical logging of hold items based on logging
    rules such as time period, update, or change.

    Act comes before Filer in .__mro__ so Act.name property is used not Filer.name

    Inherited Class Attributes:
        Registry (dict): subclass registry whose items are (name, cls) where:
                name is unique name for subclass
                cls is reference to class object
        Instances (dict): instance registry whose items are (name, instance) where:
                name is unique instance name and instance is instance reference
        Index (int): default naming index for subclass instances. Each subclass
                overrides with a subclass specific Index value to track
                subclass specific instance default names.

        HeadDirPath (str): default abs dir path head such as "/usr/local/var"
        TailDirPath (str): default rel dir path tail when using head
        CleanTailDirPath (str): default rel dir path tail when creating clean
        AltHeadDirPath (str): default alt dir path head such as  "~"
                              as fallback when desired head not permitted.
        AltTailDirPath (str): default alt rel dir path tail as fallback
                              when using alt head.
        AltCleanTailDirPath (str): default alt rel path tail when creating clean
        TempHeadDir (str): default temp abs dir path head such as "/tmp"
        TempPrefix (str): default rel dir path prefix when using temp head
        TempSuffix (str): default rel dir path suffix when using temp head and tail
        Perm (int): explicit default octal perms such as 0o1700
        Mode (str): open mode such as "r+"
        Fext (str): default file extension such as "text" for "fname.text"

    Class Attributes:
        ReservedTags (dict[str]): of reserved tags to protect from collision with
                                  defined parameter names that may not be used for
                                  log tags when using **kwa to provide hold
                                  log leys. Uses dict since inclusion test is
                                  faster than with list

    Inherited Attributes  see Act, File
        hold (Hold): data shared by boxwork

        name (str): overriden by .name property from Act (see name property)
        base (str): another unique path component inserted before name
        temp (bool): True means use TempHeadDir in /tmp directory
        headDirPath (str): head directory path
        path (str | None):  full directory or file path once created else None
        perm (int):  octal OS permissions for path directory and/or file
        filed (bool): True means .path ends in file.
                       False means .path ends in directory
        extensioned (bool): When not filed:
                                True means ensure .path ends with fext
                                False means do not ensure .path ends with fext
        mode (str): file open mode if filed
        fext (str): file extension if filed
        file (File | None): File instance when filed and created.
        opened (bool): True means directory created and if filed then file
                is opened. False otherwise


    Inherited Properties  see Act, File
        name (str): unique name string of instance used for registering Act
                    instance in Act registry as well providing a unique path
                    component used in file path name
        iops (dict): input-output-parameters for .act
        nabe (str): action nabe (context) for .act


    Attributes:
        started (bool): True means logging has begun with header
                        False means logging has not yet begun needs header
        first (float|None): tyme when began logging, None means not yet running
        last (float|None): tyme when last logged, None means not yet running
                      realtime equiv of last = began + (last - first)
        rule (str): condition for log to fire one of Rules
                    (once, every, span, update, change)
        span (float): tyme span seconds for periodic logging 0.0 means everytyme
        onced (bool): True means logged at least one (first) record
                      False means not yet logged one (first) record
        header (str): header for log file(s)
        rid (str):  universally unique run ID for given run of hog
        flushSpan (float): tyme span seconds between flushes (flush)
                           0.0 means everytyme
        flushForce (bool): True means force flush on every log
                               False means only flush at appropriate times
        flushLast (float|None): tyme last flushed, None means not yet running
        cycleCount (int): number of cycled logs, 0 means do not cycle (count)
        cyclePeriod (float): min tyme span for cycling logs (cycle). 0.0 means
                           cycle based on cycleHigh not tyme span. One of
                           cycleSpan or cycleHigh must be non zero
        cycleSize (int): maximum size in bytes allowed for each cycled log
                         0 means no maximum. One of cycleSpan or cycleHigh must
                         be non zero
        cyclePaths (list[str]): paths for cycled logs
        cycleLast (float|None): tyme last cycled. None means not yet running
        hits (dict): hold items to log. Item label is log header tag
                      Item value is hold key that provides value to log
        marks (dict): tyme or value tuples marks of hold items logged with
                      updated or changed rule. Label is hold key.
        activeKey (str|None): hold key to active box name of boxer given by iops
                              None otherwise
        tockKey (str|None): hold key to tock value of boxer given by iops
                            None otherwise
        tockKey (str|None): hold key to tyme value of boxer given by iops
                            None otherwise


    Hidden:
        _name (str): unique name of instance for .name property
        _iopts (dict): input-output-paramters for .act for .iops property
        _nabe (str): action nabe (context) for .act for .nabe property


    As ActBase subclass that is managed inside boxwork, the Hog is created and
    inited by Boxer.make which is run in the enter context of the BoxerDoer.
    So opening file during init is compatible as its in the enter context of the
    its Doer even though its not in the endo context of its box.
    It still needs to be closed. Unlike the hold subery the Boxer doesn't know
    about any Hogs runing as acts inside its boxes so can't close in its BoxerDoer
    exit context. So much be closed inside with a close do act

    Since filed it should reopen without truncating so does not overwrite
    existing log file of same name so uses mode = 'a+'.
    Need to test reopen logic
    Always rewrites header on first log after restart which may have changed
    log behavior so header demarcation enables recognition of log parameters.
    Log header includes rid (unique run id) and datatime stamp so can always
    uniquely match a given run data to header even when reusing same log file.
    This is most important when rotating logs.
    Because each individual log record after header always starts with tyme as
    floating point decimal, processor can find header demarcations interior to
    log file not merely at start.
    """
    ReservedTags = dict(name=True, iops=True, nabe=True, hold=True, base=True,
                 temp=True, headDirPath=True, perm=True, reopen=True,
                 clear=True, reuse=True, clean=True, filed=True,
                 extensioned=True, )

    Proem = "__"  # prepended to rid (run ID) in header to ensure 24 bit boundary alignment


    def __init__(self, iops=None, nabe=Nabes.afdo, base="", filed=True,
                       extensioned=True, mode='a+', fext="hog", reuse=True,
                       rid=None, rule=Rules.every, span=0.0,
                       flushSpan=60.0, flushForce=False,
                       cycleCount=0, cycleSpan=0.0, cycleSize=0,
                       hits=None, **kwa):
        """Initialize instance.

        Inherited Parameters:
            name (str|None): unique name of this instance. When None then
                generate name from .Index. Used for .name property which is
                 used for registering Act instance in Act registry as well
                 providing a unique path component used in file path name.
                 When system employs more than one installation, name allows
                 differentiating each installation by name
            iops (dict|None): input-output-parameters for .act. When None then
                set to empty dict.
            nabe (str): action nabe (context) for act. default is "endo"
            hold (None|Hold): data shared across boxwork

            base (str): optional directory path segment inserted before name
                that allows further differentation with a hierarchy. "" means
                optional.
            temp (bool): assign to .temp
                True then open in temporary directory, clear on close
                Otherwise then open persistent directory, do not clear on close
            headDirPath (str): optional head directory pathname for main database
                Default .HeadDirPath
            perm (int): optional numeric os dir permissions for database
                directory and database files. Default .DirMode
            reopen (bool): True (re)open with this init
                           False not (re)open with this init but later (default)
            clear (bool): True means remove directory upon close when reopening
                          False means do not remove directory upon close when reopening
            reuse (bool): True means reuse self.path if already exists
                          False means do not reuse but remake self.path
            clean (bool): True means path uses clean tail variant
                             False means path uses normal tail variant
            filed (bool): True means .path is file path not directory path
                          False means .path is directiory path not file path
            extensioned (bool): When not filed:
                                True means ensure .path ends with fext
                                False means do not ensure .path ends with fext
            mode (str): File open mode when filed
            fext (str): File extension when filed or extensioned

        Parameters:
            rid (str|None):  universally unique run ID for given run of hog
                             None means create one using uuid lib
            rule (str|None): condition for log to fire, one of Rules default every
                            (once, every, span, update, change)
            span (float): periodic tyme span seconds when rule is spanned
                          0.0 means every tyme
            flushSpan (float): flush tyme span seconds, tyme between flushes
                          0.0 means every tyme
            flushForce (bool): True means force flush on every log
                               False means only flush at appropriate times
            cycleCount (int): number of cycled logs, 0 means do not cycle
            cycleSapn (float): cycle tyme span, tyme between log cycles
            cycleSize (int): maximum size in bytes allowed for each cycled log
                       0 means no maximum
            hits (None|dict): hold items to log. Item label is log header tag
                      Item value is hold key that provides value to log
                    None means use unreserved items in **kwa wrt .ReservedTags


        When made (created and inited) by boxer.do then have "_boxer" and
        "_box" keys in self.iops = dict(_boxer=self.name, _box=m.box.name, **iops)

        """
        if not base:
            if iops and "_boxer" in iops:  # '_boxer' in iops when made by boxer
                base = iops['_boxer']  # set base to boxer.name

        super(Hog, self).__init__(iops=iops,
                                  nabe=nabe,
                                  base=base,
                                  filed=filed,
                                  extensioned=extensioned,
                                  mode=mode,
                                  fext=fext,
                                  reuse=reuse,
                                  **kwa)


        self.started = False
        self.first = None
        self.last = None
        self.rule = rule  # maybe should make property so readonly after init
        self.span = span
        self.onced = False
        self.rid = rid
        self.stamp = '' # need to init
        self.header = ''  # need to init
        self.flushSpan = flushSpan
        self.flushForce = flushForce
        self.flushLast = None

        if cycleCount and cycleSpan == 0.0 and cycleSize == 0:
            raise HierError(f"For non-zero count one of {cycleSpan=} or "
                            f"{cycleSize=} must be non-zero")

        self.cycleCount = max(min(cycleCount, 999), 0)
        self.cycleSpan = cycleSpan
        self.cycleSize = cycleSize
        self.cyclePaths = []  # need to init
        self.cycleLast = None

        self.activeKey = None
        self.tockKey = None
        self.tymeKey = None
        if "_boxer" in self.iops:  # assign keys for boxer box state
            boxerName = self.iops["_boxer"]
            self.activeKey = self.hold.tokey(("", "boxer", boxerName, "active"))
            self.tockKey = self.hold.tokey(("", "boxer", boxerName, "tock"))
            self.tymeKey = self.hold.tokey(("", "boxer", boxerName, "tyme"))

        if hits is None:
            hits = {}
            for tag, val in kwa.items():
                if tag not in self.ReservedTags:
                    hits[tag] = val  # value is key of hold to log

        self.hits = hits
        self.marks = {}

        if self.cycleCount > 0:
            self.cyclePaths = []
            for k in range(1, self.cycleCount + 1):
                root, ext = os.path.splitext(self.path)
                path = f"{root}_{k:03}{ext}"  # ext includes leading dot
                self.cyclePaths.append(path)
                # trial open to ensure can make
                try:
                    file = ocfn(path, 'r')  # do not truncate in case reusing
                except OSError as ex:
                    raise HierError("Failed making cycle paths") from ex

                file.close()


    def act(self, **iops):
        """Act called by ActBase.

        Parameters:
            iops (dict): input/output parms, same as self.iops. Puts **iops in
                         local scope in case act compliles exec/eval str

        When made by boxer.do then have "_boxer" and "_box" keys in self.iops
           iops = dict(_boxer=self.name, _box=m.box.name, **iops)
        """
        if not self.started:

            if self.rid is None:
                # hog id uuid for this run (not iteration)
                # create B64 version of uuid with stripped trailing pad bytes
                uid = encodeB64(bytes.fromhex(uuid.uuid1().hex))[:-2].decode()
                #self.rid = self.name + "_" + uid
                self.rid = self.Proem + uid  # same as CESR salt
            self.stamp = timing.nowIso8601()  # current real datetime as ISO8601 string

            metaLine = (f"rid\tbase\tname\tstamp\trule\tcount\n")

            metaValLine = (f"{self.rid}\t{self.base}\t{self.name}"
                           f"\t{self.stamp}\t{self.rule}\t{self.cycleCount}\n")

            if self.tymeKey and self.tymeKey in self.hold:  # need tyme for logging
                hits = dict(tyme=self.tymeKey)  # logging tyme
                for tag, key in self.hits.items():  # copy valid .hits
                    if key in self.hold:  # invalid hold key
                        hits[tag] = key  # make copy in order

            else: # need tyme in order to log anything with respect to tyme
                hits = {}  # nothing to log

            self.hits = hits

            if len(self.hits) == 1:  # only tyme so default to boxer state
                if self.activeKey and self.activeKey in self.hold:
                    self.hits["active"] = self.activeKey
                if self.tockKey and self.tockKey in self.hold:
                    self.hits["tock"] = self.tockKey

            tagKeyLine = '\t'.join(f"{tag}.key" for tag in self.hits.keys()) + "\n"
            keyLine = '\t'.join(key for key in self.hits.values()) + "\n"

            # need to expand tags for hits with vector bags in hold
            tagValLine = ('\t'.join(f"{tag}.{fld}" for tag, key in self.hits.items()
                                    for fld in self.hold[key]._names) + "\n")

            self.header = metaLine + metaValLine + tagKeyLine + keyLine + tagValLine
            self.file.write(self.header)
            self.started = True

        tyme = self.hold[self.hits["tyme"]].value if self.hits else None

        if not self.onced:
            self.first = tyme
            # always flush on first write to ensure header synced on disk
            self.log(self.record(), tyme, force=True)
            if self.hits:
                for tag, key in self.hits.items():
                    if tag != "tyme":  # do not mark tyme hold
                        if self.rule == Rules.update:  # create mark
                            self.marks[key] = self.hold[key]._tyme
                        elif self.rule == Rules.change:  # create mark
                            self.marks[key] = self.hold[key]._astuple()

            self.onced = True
        else:
            match self.rule:
                case Rules.every:
                    self.log(self.record(), tyme)
                case Rules.span:
                    if tyme is not None and tyme - self.last >= self.span:
                        self.log(self.record(), tyme)

                case Rules.update:
                    if tyme is not None:
                        updated = False
                        for key, mark in self.marks.items():  # marked tyme
                            holdTyme = self.hold[key]._tyme
                            if holdTyme > mark:  # updated since marked
                                self.marks[key] = holdTyme
                                updated = True
                        if updated:
                            self.log(self.record(), tyme)

                case Rules.change:
                    if tyme is not None:
                        changed = False
                        for key, mark in self.marks.items():  # marked value tuple
                            holdValue = self.hold[key]._astuple()
                            if holdValue != mark:  # changed since marked
                                self.marks[key] = holdValue
                                changed = True
                        if changed:
                            self.log(self.record(), tyme)

                case _:
                    pass

        return iops


    def log(self, record, tyme, force=False):
        """Write one record to file and flush when indicated

        Parameters:
            record (str):  one line of tab delimited newline ended values
            tyme (float):  tyme of log record
            force (bool): True means force flush even when not flushSpan elapsed
                          False means do not force flush only if flushSpan elapsed
        """
        self.file.write(record)
        self.last = tyme

        if force or self.flushForce or (tyme - self.flushLast) >= self.flushSpan:
            self.flush()
            self.flushLast = tyme

        if self.cycleCount:
            cycled = False
            if self.cycleSpan:
                if self.cycleLast is None:
                    delta = tyme - self.first
                else:
                    delta = tyme - self.cycleLast

                if delta >= self.cycleSpan:
                    self.cycle(tyme=tyme)
                    cycled = True

            if self.cycleSize and not cycled:
                try:
                    size = os.path.getsize(self.path)
                except OSError as ex:
                    pass
                else:
                    if size >= self.cycleSize:
                        self.cycle(tyme=tyme)


    def record(self):
        """Generate on record line .hits values from .hold

        Returns:
           record (str): one newline delimited line of tab delimited values
                         from .hits. Each hit time value is key into hold
                         vector holds each get entry in record per field


        """
        # hit values are hold keys
        if self.hits:
            return ('\t'.join(f"{self.hold[key][fld]}"
                                for key in self.hits.values()
                                    for fld in self.hold[key]._names) + "\n")
        else:
            return ""


    def cycle(self, tyme):
        """Cycle log files

        Parameters:
            tyme (float): current tyme of cycle
        """
        self.flush
        cycles = deque([self.path])
        cycles.extend(self.cyclePaths)

        cycled = False  # if cycling is successful
        new = cycles.pop()
        while cycles:
            old = cycles.pop()
            try:
                if old == self.path:
                    self.close()
                os.replace(old, new)  # rename old to new thereby clobbering old
                cycled = True
            except OSError as ex:
                cycled = False  # failed to cycle so do not clobber self.file
                break
            new = old  # old path is now free to use

        if not cycled:  # reopen cleanly just in case
            self.reopen(reuse=True)  # reopen reuse, .mode is "a+" so saves
        else:  # all cycled so recreate self.file
            self.reopen() # not reuse so recreates empty

        self.file.write(self.header)  # rewrite header
        self.flush()
        self.flushLast = tyme
        self.cycleLast = tyme


@contextmanager
def openHog(cls=None, name=None, temp=True, reopen=True, clear=False, **kwa):
    """Context manager wrapper Hog instances for managing a filesystem directory
    and or files in a directory.

    Defaults to using temporary directory path.
    Context 'with' statements call .close on exit of 'with' block

    Parameters:
        cls is Class instance of subclass instance
        name is str name of Filer instance path part so can have multiple Filers
             at different paths that each use different dirs or files
        temp is Boolean, True means open in temporary directory, clear on close
                Otherwise open in persistent directory, do not clear on close
        reopen (bool): True (re)open with this init
                           False not (re)open with this init but later (default)
        clear (bool): True means remove directory upon close when reopening
                          False means do not remove directory upon close when reopening
    See hogging.Hog for other keyword parameter passthroughs

    Usage:

    with openHog(name="bob") as hog:

    with openHog(name="eve", cls=HogSubClass) as hog:

    """
    hog = None
    if cls is None:
        cls = Hog
    try:
        hog = cls(name=name, temp=temp, reopen=reopen, clear=clear, **kwa)
        yield hog

    finally:
        if hog:
            hog.close(clear=hog.temp or clear)  # clears if hog.temp



class HogDoer(Doer):
    """
    Basic Hog Doer

    Attributes:
        done (bool): completion state:
            True means completed
            Otherwise incomplete. Incompletion maybe due to close or abort.
        hog (Hog): instance

    Properties:
        tyme (float): relative cycle time of associated Tymist .tyme obtained
            via injected .tymth function wrapper closure.
        tymth (func): closure returned by Tymist .tymeth() method.
            When .tymth is called it returns associated Tymist .tyme.
            .tymth provides injected dependency on Tymist tyme base.
        tock (float)): desired time in seconds between runs or until next run,
                 non negative, zero means run asap

    """

    def __init__(self, hog, **kwa):
        """
        Parameters:
           tymist (Tymist): instance
           tock (float): initial value of .tock in seconds
           hog (Hog): instance
        """
        super(HogDoer, self).__init__(**kwa)
        self.hog = hog


    def enter(self, *, temp=None):
        """Do 'enter' context actions. Override in subclass. Not a generator method.
        Set up resources. Comparable to context manager enter.

        Parameters:
            temp (bool | None): True means use temporary file resources if any
                                None means ignore parameter value. Use self.temp

        Inject temp or self.temp into file resources here if any
        """
        # inject temp into file resources here if any
        if not self.hog.opened:
            self.hog.reopen(temp=temp)


    def exit(self):
        """"""
        self.hog.close(clear=self.hog.temp)



# -*- encoding: utf-8 -*-
"""hio.hold.holding Module

Provides durable data storage with LMDB

"""
from __future__ import annotations  # so type hints of classes get resolved later

import os
import platform
import stat
import tempfile
from collections.abc import Mapping
from contextlib import contextmanager
from typing import Any

import lmdb
from ordered_set import OrderedSet as oset

from hio import HierError

from ...help import NonStringIterable, Mine
from ..doing import Doer
from .canning import CanDom
from .durqing import Durq
from .dusqing import Dusq




class Hold(Mine):
    """Hold is Mine subclass that on writes intercepts key and keys and then
    also saves updates value in durable storage for value object CanDom and Durq
    serializations

    Hold has special item '_hold_subery' whose value is Subery instance or None.
    Read access is provided via the property .subery

    Subery attributes are the subdbs:
    .cans
    .dsqs

    Hold injects reference to one of these based on the type of value and the Hold
    key  into ._sdb and ._key.  When not a durable value then no injection.
    This allows the value then to read/write to its ._sdb at ._key the fields
    of the Can when Can or Durq when Durq


    Properties:
        subery (None|Subery): gets instance from item at '_hold_subery' or None

    Methods:
        inject(self, key, val): injects .cans into val._sdb and key into val._key


    """
    def __init__(self, *pa, **kwa):
        """Convert keys that are tuples when positional argument is Iterable or
        Mapping to '.' joined strings

        dict __init__ signature options are:
            dict(**kwa)
            dict(mapping, **kwa)
            dict(iterable, **kwa)
        dict.update has same call signature
            d.update({"a": 5, "b": 2,}, c=3 , d=4)
        """
        self._hold_subery = None
        super(Hold, self).__init__(*pa, **kwa)


    def __setitem__(self, k, v):  # __setattr__ calls __setitem__
        k = self.tokey(k)  # get key to inject
        result = super(Hold, self).__setitem__(k, v)
        self.inject(k, v)
        return result


    def update(self, *pa, **kwa):
        """Convert keys that are tuples when positional argument is Iterable or
        Mapping to '.' joined strings

        dict __init__ signature options are:
            dict(**kwa)
            dict(mapping, **kwa)
            dict(iterable, **kwa)
        dict.update has same call signature
            d.update({"a": 5, "b": 2,}, c=3 , d=4)

        """
        if len(pa) > 1:
            raise TypeError(f"Expected 1 positional argument got {len(pa)}.")

        if pa:
            di = pa[0]
            if isinstance(di, Mapping):
                rd = {}
                for k, v in di.items():
                    rd[self.tokey(k)] = v
                super(Hold, self).update(rd, **kwa)
                for k, v in rd.items():
                    self.inject(k, v)

            elif isinstance(di, NonStringIterable):
                ri = []
                for k, v in di:
                    ri.append((self.tokey(k), v))
                super(Hold, self).update(ri, **kwa)
                for k, v in ri:
                    self.inject(k, v)

            else:
                raise TypeError(f"Expected Mapping or NonStringIterable got "
                                f"{type(di)}.")

        else:
            super(Hold, self).update(**kwa)

        for k, v in kwa.items():
            self.inject(k, v)



    def inject(self, key, val):
        """When val is instance of CanDom, injects .tokey(key) into val._key and
        .subery.cans into val._sdb

        Parameters:
            key (str): for item
            val (Any|CanDom): for item. When instance subclass of CanDom then
                inject to ._key and ._sdb
        """
        if isinstance(val, CanDom):
            val._key = key
            val._sdb = self.subery.cans if self.subery else None
            val._sync()  # attempt to sync with sdb at key if any
        elif isinstance(val, Durq):
            val._key = key
            val._sdb = self.subery.drqs if self.subery else None
            val.sync()  # attempt to sync with sdb at key if any
        elif isinstance(val, Dusq):
            val._key = key
            val._sdb = self.subery.dsqs if self.subery else None
            val.sync()  # attempt to sync with sdb at key if any


    @property
    def subery(self):
        """Gets value of special item '_hold_subery'

        Returns:
            subery (None|Subery): instance or None
        """
        return self["_hold_subery"]



